"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The RAVA driver implements the code for communicating with an RAVA device
running the RAVA firmware. The computer running the driver assumes the leader's
role, sending command requests and reading the data replies.

The RAVA_RNG class enables the request of pulse counts, random bits, random
bytes, and random numbers (integers and floats). Additionally, it establishes
the circuit's basic functionality encompassing key modules such as EEPROM, PWM,
heath tests, peripherals, and interfaces.

The functions that provide access to RAVA's functionality are prefixed with
"snd" and "get". "Snd" commands are unidirectional and do not expect a device's
response. Conversely, "get" commands are bidirectional, where the driver
immediately attempts to retrieve the expected information after signaling the
RAVA device.

The communication exchanges start with an 8-byte message. The first byte holds
the character '$' (00100100), signifying the message start. The second byte
encodes the command's identification code, while the subsequent bytes house the
command's specific data. The commands are sent to the RAVA device using the
snd_rava_msg() function, which generates the 8-byte information with the
pack_rava_msg() function.

The driver operates with a parallel thread running the loop_serial_listen()
function to continuously monitor RAVA responses. When a new message is detected,
it undergoes further processing within the process_serial_comm() function. This
function stores the command's variables in a queue object associated with the
respective command ID. To achieve this, the 6-byte data is transformed into the
command's specific variables using the unpack_rava_msgdata() function. These
variables are then available for retrieval by employing the get_queue_data()
function.

The queue-based design not only mitigates read-ordering conflicts but also
serves as the foundation for the asynchronous driver version, which uses asyncio
queues and implements get_queue_data() as an async function. This capability is
realized in the RAVA_RNG_AIO class.

The cable disconnection of an operational device is detected by the serial
functions. In response, the close() method is invoked. This method halts the
serial listen loop, closes the serial connection, and triggers the registered
device_close callback function.
"""

import logging
import struct
import time
import threading
import queue
import weakref

import serial
from serial.tools.list_ports import comports
import numpy as np

from rng_rava.rava_defs import *


### FUNCTIONS

def find_rava_sns(usb_vid=RAVA_USB_VID, usb_pid=RAVA_USB_PID):
    sns = [port_info.serial_number for port_info in comports()
            if (port_info.vid == usb_vid and port_info.pid == usb_pid)]
    sns.sort()
    return sns


def find_rava_port(serial_number):
    if isinstance(serial_number, bytes):
        serial_number = serial_number.decode()
    ports = [port_info.device for port_info in comports()
             if port_info.serial_number == serial_number]
    return ports[0] if len(ports) else None


def find_usb_info(port):
    return [(port_info.product, port_info.serial_number, port_info.vid, port_info.pid) for port_info in comports()
            if (port_info.device == port)][0]


def version_a_greater_b(v_a, v_b):
    v_ints_a = [int(i) for i in v_a.split('.')]
    v_ints_b = [int(i) for i in v_b.split('.')]

    # v_a > v_b?
    for i in range(len(v_ints_a)):
        if v_ints_a[i] > v_ints_b[i]:
            return True

    return False


def bytes_to_list(bytes_data, num_type='B'):
    int_size = len(bytes_data)
    if num_type.lower() == 'h':
        int_size = int_size // 2
    elif num_type.lower() in ['i', 'l', 'f']:
        int_size = int_size // 4
    elif num_type.lower() in ['q', 'd']:
        int_size = int_size // 8

    unpack_str = '<{}{}'.format(int_size, num_type)
    return list(struct.unpack(unpack_str, bytes_data))


def bytes_to_array(bytes_data, array_dtype=np.uint8):
    # Force little-endian
    array_dtype = np.dtype(array_dtype).newbyteorder('<')

    return np.frombuffer(bytes_data, dtype=array_dtype)


def print_health_startup_results(test_success, test_vars_dict):
    success_str = 'Success' if test_success else 'Failed'

    pc_result, pc_a, pc_b, pc_min = test_vars_dict['pulse_count']
    pc_diff_result, pc_diff_a, pc_diff_b, pc_diff_min = test_vars_dict['pulse_count_diff']
    bias_result, bias_a, bias_b, bias_abs_treshold = test_vars_dict['bit_bias']
    chisq_result, chisq_a, chisq_b, chisq_max_treshold = test_vars_dict['byte_bias']

    lg = logging.getLogger('rava')
    lg.info('Startup Health Tests: {}'.format(success_str) + \
            '\n{}Pulse Count: {}'.format(LOG_FILL, bool(pc_result)) + \
            '\n{}  pc_a={:.2f}, pc_b={:.2f}, pc_tresh={:.2f}'\
                .format(LOG_FILL, pc_a, pc_b, pc_min) + \
            '\n{}Pulse Count Difference: {}'.format(LOG_FILL, bool(pc_diff_result)) + \
            '\n{}  pc_diff_a={:.2f}, pc_diff_b={:.2f}, pc_diff_tresh={:.2f}'\
                .format(LOG_FILL, pc_diff_a, pc_diff_b, pc_diff_min) + \
            '\n{}Bit bias: {}'.format(LOG_FILL, bool(bias_result)) + \
            '\n{}  bias_a={:.4f}, bias_b={:.4f}, bias_tresh={:.2f}'\
                .format(LOG_FILL, bias_a, bias_b, bias_abs_treshold) + \
            '\n{}Byte bias: {}'.format(LOG_FILL, bool(chisq_result)) + \
            '\n{}  chisq_a={:.4f}, chisq_b={:.4f}, chisq_tresh={:.2f}'
                .format(LOG_FILL, chisq_a, chisq_b, chisq_max_treshold))


### RAVA_RNG

class RAVA_RNG:

    rava_instances = weakref.WeakSet()

    def __init__(self, dev_name='RAVA_RNG'):
        self.dev_name = dev_name
        self.queue_type = queue.Queue

        # Logging
        self.log_setup()

        # Debug
        self.lg.debug('> {} INIT'.format(self.dev_name))

        # Variables defined upon connection
        self.dev_serial_number = ''
        self.firmware_dict = None
        self.dev_usb_port = ''
        self.dev_usb_name = ''
        self.dev_usb_vid = None
        self.dev_usb_pid = None

        self.health_startup_enabled = None
        self.health_startup_success = None
        self.health_continuous_enabled = None
        self.led_enabled = None
        self.lamp_enabled = None
        self.peripherals_enabled = None

        # Serial variables
        self.serial = serial.Serial()
        self.serial_connected = threading.Event() # Used by loop_serial_listen()
        self.serial_read_lock = threading.Lock()
        self.serial_write_lock = threading.Lock()
        self.serial_data = {}
        self.serial_listen_thread = None

        # Byte streaming variables
        self.rng_streaming = False

        # Callback functions
        self.cbkfcn_device_close = lambda: None
        self.cbkfcn_rng_stream_data_available = lambda: None
        self.cbkfcn_d2_input_capture_available = lambda: None

        # Finalize any previous RAVA instance
        for rava in self.rava_instances.copy():
            rava.close()
            self.rava_instances.remove(rava)

        # Add new instance to weakref list
        self.rava_instances.add(self)


    def __del__(self):
        # Debug
        self.lg.debug('> {} DEL'.format(self.dev_name))


    def connect(self, serial_number):
        # Debug
        self.lg.debug('> {} CONNECT'.format(self.dev_name))

        # Find serial port
        port = find_rava_port(serial_number)
        if port is None:
            self.lg.error('{} Connect: No device found with SN {}'.format(self.dev_name, serial_number))
            return False

        # Open serial connection
        if not self.open_serial(port):
            return False

        # Reset serial data queues
        self.init_queue_data()

        # Start listening for serial commands
        self.serial_listen_thread = threading.Thread(target=self.loop_serial_listen)
        self.serial_listen_thread.start()

        # Stop any active RNG byte stream
        self.snd_rng_byte_stream_stop()

        # Request firmware info
        self.firmware_dict = self.get_eeprom_firmware()
        self.firmware_version =  self.firmware_dict['version']
        self.health_startup_enabled = self.firmware_dict['health_startup_enabled']
        self.health_continuous_enabled = self.firmware_dict['health_continuous_enabled']
        self.led_enabled = self.firmware_dict['led_enabled']
        self.lamp_enabled = self.firmware_dict['lamp_enabled']
        self.peripherals_enabled = self.firmware_dict['peripherals_enabled']

        # Check Firmware version
        if version_a_greater_b(FIRMWARE_MIN_VERSION, self.firmware_version):
            self.lg.error('{} Connect: Firmware version v{} is incompatible with the driver'.format(self.dev_name, self.firmware_version))
            return False

        # Print connection info
        self.lg.info('{} Connect: Success'
                '\n{}  {}, Firmware v{}, SN={}, at {}'
                .format(self.dev_name, LOG_FILL,
                        self.dev_usb_name, self.firmware_version, self.dev_serial_number, self.serial.port))

        # Request Health startup info
        if self.health_startup_enabled:
            self.health_startup_success, test_vars = self.get_health_startup_results()

            # Print test info
            print_health_startup_results(self.health_startup_success, test_vars)

            # Error? Users have then a limited command variety (see Firmware code)
            if not self.health_startup_success:
                self.lg.error('{} Connect: Startup tests failed'.format(self.dev_name))
                return False

        # Reset Health continuous erros
        self.get_health_continuous_errors()

        return True


    def connected(self):
        return self.serial.is_open


    def close(self):
        # Debug
        self.lg.debug('> {} CLOSE'.format(self.dev_name))

        # Stop loop_serial_listen
        self.serial_connected.clear()

        # Close serial connection
        self.serial.close()

        # Callback
        try:
            self.cbkfcn_device_close()
        except:
            pass


    ## LOGGING

    def log_setup(self):
        self.lg = logging.getLogger('rava')

        # RAVA Logger 1st call
        if not self.lg.hasHandlers():

            # Create handler
            lg_fmt = logging.Formatter(fmt='%(asctime)s %(levelname)-7s %(message)s', datefmt='%H:%M:%S')
            lg_sh = logging.StreamHandler()
            lg_sh.setFormatter(lg_fmt)
            self.lg.addHandler(lg_sh)

            # Set defaulf level to INFO
            self.log_level('INFO')


    def log_level(self, level_str, level_id=None):
        if level_id:
            level = level_id
        else:
            try:
                level = logging.getLevelNamesMapping()[level_str]
            except:
                level = logging.INFO

        self.lg.setLevel(level)


    ## QUEUE

    def init_queue_data(self):
        # Create one queue for each command
        self.serial_data = {}
        for comm in D_DEV_COMM.keys():
            self.serial_data[comm] = self.queue_type()


    def put_queue_data(self, comm, value, comm_ext_id=0):
        # Check key
        if comm not in self.serial_data:
            self.lg.error('{} Data: Unknown comm {}'.format(self.dev_name, comm))
            return False

        # Extended key
        if comm_ext_id:
            comm_ext = '{}_{}'.format(comm, comm_ext_id)

            # Queue exists?
            if comm_ext not in self.serial_data:
                self.serial_data[comm_ext] = self.queue_type()
        else:
            comm_ext = comm

        # Try writing to queue
        try:
            self.serial_data[comm_ext].put(value)
            return True

        except queue.Full:
            self.lg.error('{} Data: {} Queue full'.format(self.dev_name, comm_ext))
            return False


    def get_queue_data(self, comm, comm_ext_id=0, timeout=GET_TIMEOUT_S):
        # Check key
        if comm not in self.serial_data:
            self.lg.error('{} Data: Unknown comm {}'.format(self.dev_name, comm))
            return None

        # Extended key
        if comm_ext_id:
            comm_ext = '{}_{}'.format(comm, comm_ext_id)

            # Queue exists?
            if comm_ext not in self.serial_data:
                self.serial_data[comm_ext] = self.queue_type()
        else:
            comm_ext = comm

        # Try reading from queue
        try:
            return self.serial_data[comm_ext].get(timeout=timeout)

        except queue.Empty:
            self.lg.error('{} Data: Timeout retrieving {}'.format(self.dev_name, comm_ext))
            return None


    ## SERIAL

    def open_serial(self, port):
        # Set serial port
        self.serial.port = port

        # Open serial connection
        try:
            self.serial.open()
            self.serial_connected.set()

            # Get USB parameters
            self.dev_usb_port = port
            self.dev_usb_name, self.dev_serial_number, self.dev_usb_vid, self.dev_usb_pid = find_usb_info(port)
            return True

        except Exception as err:
            self.lg.error('{} Serial: Failed opening {}'
                     '\n{}  {} - {}'
                     .format(self.dev_name, port, LOG_FILL, type(err).__name__, err))
            return False


    def inwaiting_serial(self):
        # Read in_waiting
        try:
            return self.serial.in_waiting

        except Exception as err:
            self.lg.error('{} Serial: Failed reading in_waiting'
                     '\n{}  {} - {}'
                     .format(self.dev_name, LOG_FILL, type(err).__name__, err))

            # Close device
            self.close()
            return None


    def read_serial(self, n_bytes):
        # Read serial
        try:
            with self.serial_read_lock:
                data = self.serial.read(n_bytes)
            return data

        except Exception as err:
            self.lg.error('{} Serial: Failed reading'
                     '\n{}  {} - {}'
                     .format(self.dev_name, LOG_FILL, type(err).__name__, err))

            # Close device
            self.close()
            return None


    def write_serial(self, comm_bytes):
        # Write serial
        try:
            with self.serial_write_lock:
                self.serial.write(comm_bytes)
            return True

        except Exception as err:
            self.lg.error('{} Serial: Failed writing'
                    '\n{}  {} - {}'
                    .format(self.dev_name, LOG_FILL, type(err).__name__, err))

            # Close device
            self.close()
            return False


    def loop_serial_listen(self):
        # Debug
        self.lg.debug('> {} LOOP SERIAL LISTEN'.format(self.dev_name))

        # Loop while connected
        while self.serial_connected.is_set():

            # Command available?
            comm_inwaiting = self.inwaiting_serial()
            if comm_inwaiting is None:
                continue # Disconnected

            if comm_inwaiting > 0:

                # Read command starting char
                comm_start = self.read_serial(1)
                if comm_start is None:
                    continue # Disconnected

                # Starts with $?
                if comm_start == COMM_MSG_START:

                    # Read remaining command bytes
                    comm_msg = self.read_serial(COMM_MSG_LEN-1)
                    if comm_msg is None:
                        continue # Disconnected

                    comm_id = comm_msg[0]
                    comm_data = comm_msg[1:]

                    # Known command id?
                    if comm_id in D_DEV_COMM_INV:

                        # Debug
                        self.lg.debug('> COMM RCV {}'.format([D_DEV_COMM_INV[comm_id], *[c for c in comm_data]]))

                        # Process Command
                        try:
                            self.process_serial_comm(comm_id, comm_data)

                        except Exception as err:
                            self.lg.error('{} Serial Listen Loop: Error processing command_id {}'
                                        '\n{}  {} - {}'
                                        .format(self.dev_name, comm_id, LOG_FILL, type(err).__name__, err))

                            # Close device
                            self.close()

                    else:
                        self.lg.warning('{} Serial Listen Loop: Unknown command_id {}'.format(self.dev_name, comm_id))

                else:
                    self.lg.warning('{} Serial Listen Loop: Commands must start with {}'.format(self.dev_name, COMM_MSG_START))

            # The non-blocking method is prefered for finishing the thread
            # when closing the device
            else:
                time.sleep(SERIAL_LISTEN_LOOP_INTERVAL_S)


    def process_serial_comm(self, comm_id, comm_data):
        # DEVICE_SERIAL_NUMBER
        if comm_id == D_DEV_COMM['DEVICE_SERIAL_NUMBER']:
            sn_n_bytes = self.unpack_rava_msgdata(comm_data, 'B')
            sn = self.read_serial(sn_n_bytes).decode()
            self.put_queue_data('DEVICE_SERIAL_NUMBER', sn)

        # DEVICE_TEMPERATURE
        elif comm_id == D_DEV_COMM['DEVICE_TEMPERATURE']:
            # temperature
            self.put_queue_data('DEVICE_TEMPERATURE', self.unpack_rava_msgdata(comm_data, 'f'))

        # DEVICE_FREE_RAM
        elif comm_id == D_DEV_COMM['DEVICE_FREE_RAM']:
            # free_ram
            self.put_queue_data('DEVICE_FREE_RAM', self.unpack_rava_msgdata(comm_data, 'H'))

        # DEVICE_DEBUG
        elif comm_id == D_DEV_COMM['DEVICE_DEBUG']:
            debug_bytes = self.unpack_rava_msgdata(comm_data, 'BBBBBB')
            debug_ints = [x for x in debug_bytes]
            self.lg.debug('> RAVA DEBUG MSG {} {} {} {} {} {}'.format(*debug_ints))

        # EEPROM_DEVICE
        elif comm_id == D_DEV_COMM['EEPROM_DEVICE']:
            # temp_calib_slope, temp_calib_intercept
            self.put_queue_data('EEPROM_DEVICE', self.unpack_rava_msgdata(comm_data, 'Hh'))

        # EEPROM_FIRMWARE
        elif comm_id == D_DEV_COMM['EEPROM_FIRMWARE']:
            # version_major, version_minor, version_patch, modules
            self.put_queue_data('EEPROM_FIRMWARE', self.unpack_rava_msgdata(comm_data, 'BBBB'))

        # EEPROM_PWM
        elif comm_id == D_DEV_COMM['EEPROM_PWM']:
            # freq_id, duty
            self.put_queue_data('EEPROM_PWM', self.unpack_rava_msgdata(comm_data, 'BB'))

        # EEPROM_RNG
        elif comm_id == D_DEV_COMM['EEPROM_RNG']:
            # sampling_interval_us
            self.put_queue_data('EEPROM_RNG', self.unpack_rava_msgdata(comm_data, 'B'))

        # EEPROM_LED
        elif comm_id == D_DEV_COMM['EEPROM_LED']:
            # led_attached
            self.put_queue_data('EEPROM_LED', self.unpack_rava_msgdata(comm_data, '?'))

        # EEPROM_LAMP
        elif comm_id == D_DEV_COMM['EEPROM_LAMP']:
            exp_mag_smooth_n_trials, extra_n_bytes = self.unpack_rava_msgdata(comm_data, 'BB')
            extra_bytes = self.read_serial(extra_n_bytes)
            if extra_bytes is None:
                exp_dur_max_ms, exp_z_significant = None, None
            else:
                exp_dur_max_ms, exp_z_significant = struct.unpack('<Lf', extra_bytes)
            self.put_queue_data('EEPROM_LAMP', (exp_dur_max_ms, exp_z_significant, exp_mag_smooth_n_trials))

        # PWM_SETUP
        elif comm_id == D_DEV_COMM['PWM_SETUP']:
            # freq_id, duty
            self.put_queue_data('PWM_SETUP', self.unpack_rava_msgdata(comm_data, 'BB'))

        # RNG_SETUP
        elif comm_id == D_DEV_COMM['RNG_SETUP']:
            # sampling_interval_us
            self.put_queue_data('RNG_SETUP', self.unpack_rava_msgdata(comm_data, 'B'))

        # RNG_PULSE_COUNTS
        elif comm_id == D_DEV_COMM['RNG_PULSE_COUNTS']:
            n_counts = self.unpack_rava_msgdata(comm_data, 'H')
            counts_bytes = self.read_serial(2 * n_counts)
            if counts_bytes is None:
                counts_bytes_a = None
                counts_bytes_b = None
            else:
                counts_bytes_a = counts_bytes[::2]
                counts_bytes_b = counts_bytes[1::2]
            self.put_queue_data('RNG_PULSE_COUNTS', (counts_bytes_a, counts_bytes_b))

        # RNG_BITS
        elif comm_id == D_DEV_COMM['RNG_BITS']:
            bit_type, bit_a, bit_b = self.unpack_rava_msgdata(comm_data, 'BBB')
            if bit_type == D_RNG_BIT_SRC['AB']:
                self.put_queue_data('RNG_BITS', (bit_type, bit_a, bit_b))
            else:
                self.put_queue_data('RNG_BITS', (bit_type, bit_a))

        # RNG_BYTES
        elif comm_id == D_DEV_COMM['RNG_BYTES']:
            n_bytes, request_id = self.unpack_rava_msgdata(comm_data, 'HB')
            rng_bytes = self.read_serial(n_bytes * 2)
            if rng_bytes is None:
                rng_bytes_a = None
                rng_bytes_b = None
            else:
                rng_bytes_a = rng_bytes[::2]
                rng_bytes_b = rng_bytes[1::2]
            self.put_queue_data('RNG_BYTES', (rng_bytes_a, rng_bytes_b), comm_ext_id=request_id)

        # RNG_STREAM_BYTES
        elif comm_id == D_DEV_COMM['RNG_STREAM_BYTES']:
            n_bytes = self.unpack_rava_msgdata(comm_data, 'H')
            rng_bytes = self.read_serial(n_bytes * 2)
            if self.rng_streaming:
                if rng_bytes is None:
                    rng_bytes_a = None
                    rng_bytes_b = None
                else:
                    rng_bytes_a = rng_bytes[::2]
                    rng_bytes_b = rng_bytes[1::2]
                self.put_queue_data('RNG_STREAM_BYTES', (rng_bytes_a, rng_bytes_b))
                # Callback
                try:
                    self.cbkfcn_rng_stream_data_available()
                except:
                    pass

        # RNG_STREAM_STATUS
        elif comm_id == D_DEV_COMM['RNG_STREAM_STATUS']:
            # stream_status
            self.put_queue_data('RNG_STREAM_STATUS', self.unpack_rava_msgdata(comm_data, '?'))

        # RNG_INT8S
        elif comm_id == D_DEV_COMM['RNG_INT8S']:
            n_ints = self.unpack_rava_msgdata(comm_data, 'H')
            # ints_bytes
            self.put_queue_data('RNG_INT8S', self.read_serial(n_ints))

        # RNG_INT16S
        elif comm_id == D_DEV_COMM['RNG_INT16S']:
            n_ints = self.unpack_rava_msgdata(comm_data, 'H')
            # ints_bytes
            self.put_queue_data('RNG_INT16S', self.read_serial(n_ints * 2))

        # RNG_FLOATS
        elif comm_id == D_DEV_COMM['RNG_FLOATS']:
            n_floats = self.unpack_rava_msgdata(comm_data, 'H')
            # ints_bytes
            self.put_queue_data('RNG_FLOATS', self.read_serial(n_floats * 4))

        # HEALTH_STARTUP_RESULTS
        elif comm_id == D_DEV_COMM['HEALTH_STARTUP_RESULTS']:
            success, pc_n_bytes, pc_diff_n_bytes, bias_bit_n_bytes, bias_byte_n_bytes = self.unpack_rava_msgdata(comm_data, '?BBBB')

            # pc_result, pc_a, pc_b, pc_min
            pc_bytes = self.read_serial(pc_n_bytes)
            if pc_bytes is None:
                pc_vars = None
            else:
                pc_vars = struct.unpack('<?fff', pc_bytes)

            # pc_diff_result, pc_diff_a, pc_diff_b, pc_avg_diff_min
            pc_diff_bytes = self.read_serial(pc_diff_n_bytes)
            if pc_diff_bytes is None:
                pc_diff_vars = None
            else:
                pc_diff_vars = struct.unpack('<?fff', pc_diff_bytes)

            # bias_result, bias_a, bias_b, bias_abs_treshold
            bias_bit_bytes = self.read_serial(bias_bit_n_bytes)
            if bias_bit_bytes is None:
                bias_bit_vars = None
            else:
                bias_bit_vars = struct.unpack('<?fff', bias_bit_bytes)

            # chisq_result, chisq_a, chisq_b, chisq_max_treshold
            bias_byte_bytes = self.read_serial(bias_byte_n_bytes)
            if bias_byte_bytes is None:
                bias_byte_vars = None
            else:
                bias_byte_vars = struct.unpack('<?fff', bias_byte_bytes)

            self.put_queue_data('HEALTH_STARTUP_RESULTS', (success, pc_vars, pc_diff_vars, bias_bit_vars, bias_byte_vars))

        # HEALTH_CONTINUOUS_ERRORS
        elif comm_id == D_DEV_COMM['HEALTH_CONTINUOUS_ERRORS']:
            n_extra_bytes = self.unpack_rava_msgdata(comm_data, 'B')
            # repetitive_count_a, repetitive_count_b, adaptative_proportion_a, adaptative_proportion_b
            health_extra_bytes = self.read_serial(n_extra_bytes)
            if health_extra_bytes is None:
                health_extra_vars = None
            else:
                health_extra_vars = struct.unpack('<HHHH', health_extra_bytes)

            self.put_queue_data('HEALTH_CONTINUOUS_ERRORS', health_extra_vars)

        # PERIPH_READ
        elif comm_id == D_DEV_COMM['PERIPH_READ']:
            # digi_state, digi_mode
            self.put_queue_data('PERIPH_READ', self.unpack_rava_msgdata(comm_data, 'BB'))

        # PERIPH_D2_TIMER3_INPUT_CAPTURE
        elif comm_id == D_DEV_COMM['PERIPH_D2_TIMER3_INPUT_CAPTURE']:
            # input_capture_interval_s
            self.put_queue_data('PERIPH_D2_TIMER3_INPUT_CAPTURE', self.unpack_rava_msgdata(comm_data, 'f'))
            # Callback
            try:
                self.cbkfcn_d2_input_capture_available()
            except:
                pass

        # PERIPH_D5_ADC
        elif comm_id == D_DEV_COMM['PERIPH_D5_ADC']:
            # adc_read_mv
            self.put_queue_data('PERIPH_D5_ADC', self.unpack_rava_msgdata(comm_data, 'f'))

        # INTERFACE_DS18B20
        elif comm_id == D_DEV_COMM['INTERFACE_DS18B20']:
            # temperature
            self.put_queue_data('INTERFACE_DS18B20', self.unpack_rava_msgdata(comm_data, 'f'))


    ## RAVA COMM

    def pack_rava_msg(self, comm_str, data_user=[], data_user_fmt=''):
        # Transform comm and variables into a 8-byte RAVA message

        # Get comm id
        if comm_str not in D_DEV_COMM:
            self.lg.error('{} MSG Pack: Unknown command {}'.format(self.dev_name, comm_str))
            return None
        comm_id = D_DEV_COMM[comm_str]

        # Check if data and data_fmt have the same size
        if len(data_user) != len(data_user_fmt):
            self.lg.error('{} MSG Pack: Data and data_fmt must have the same size'.format(self.dev_name))
            return None

        # Msg data contains (COMM_MSG_LEN - 2) bytes
        # (as the first 2 are always $ and the command id)
        data_usr_n_bytes = struct.calcsize('<' + data_user_fmt)
        data_n_bytes = COMM_MSG_LEN - 2
        if  data_usr_n_bytes > data_n_bytes:
            self.lg.error('{} MSG Pack: Data contains {} bytes - the maximum is {}'
                     .format(self.dev_name, data_usr_n_bytes, data_n_bytes))
            return None

        # Fill remaining data bytes with 0
        data_fmt = data_user_fmt + (data_n_bytes - data_usr_n_bytes) * 'B'
        data = data_user + (data_n_bytes - data_usr_n_bytes) * [0]

        # Pack rava msg
        return struct.pack('<cB' + data_fmt, COMM_MSG_START, comm_id, *data)


    def snd_rava_msg(self, comm_str, data_user=[], data_user_fmt=''):
        # Pack rava msg
        comm_bytes = self.pack_rava_msg(comm_str=comm_str, data_user=data_user, data_user_fmt=data_user_fmt)
        if comm_bytes is None:
            return False

        # Send command
        write_success = self.write_serial(comm_bytes)

        # Debug
        if write_success:
            comm_dbg = [c for c in comm_bytes][1:]
            comm_dbg[0] = D_DEV_COMM_INV[comm_dbg[0]]
            self.lg.debug('> COMM SND {}'.format(comm_dbg))

        return write_success


    def unpack_rava_msgdata(self, data, data_get_format):
        # Transform RAVA 6-byte message data into variables

        # Msg data contains (COMM_MSG_LEN - 2) bytes
        data_n_bytes = COMM_MSG_LEN - 2
        if  len(data) != data_n_bytes:
            self.lg.error('{} MSG Unpack: Data contains {} bytes - expected {}'
                     .format(self.dev_name, len(data), data_n_bytes))
            return None

        # Check data_get_fmt
        dataget_n_bytes = struct.calcsize('<' + data_get_format)
        if (dataget_n_bytes == 0) or (dataget_n_bytes > data_n_bytes):
            self.lg.error('{} MSG Unpack: Data format asks for {} bytes - expected 0 < size <= {}'
                     .format(self.dev_name, dataget_n_bytes, data_n_bytes))
            return None

        # Fill remaining data fmt bytes
        data_format = data_get_format + (data_n_bytes - dataget_n_bytes) * 'B'

        # Unpack and return data
        vars_n = len(data_get_format)
        vars = struct.unpack('<' + data_format, data)
        if vars_n == 1:
            return vars[:vars_n][0]
        else:
            return vars[:vars_n]


    ## CALLBACK REGISTER

    def cbkreg_device_close(self, fcn_device_close):
        if callable(fcn_device_close):
            self.cbkfcn_device_close = fcn_device_close
            self.lg.debug('{} Callback: Registering Device Close function to {}'
                    .format(self.dev_name, fcn_device_close.__name__))
        elif fcn_device_close is None:
            self.cbkfcn_device_close = lambda: None
            self.lg.debug('{} Callback: Unregistering Device Close function'.format(self.dev_name))
        else:
            self.lg.error('{} Callback: Provide fcn_device_close as a function or None'.format(self.dev_name))


    def cbkreg_stream_data_available(self, fcn_rng_stream_data_available):
        if callable(fcn_rng_stream_data_available):
            self.cbkfcn_rng_stream_data_available = fcn_rng_stream_data_available
            self.lg.debug('{} Callback: Registering Stream Data Available function to {}'
                    .format(self.dev_name, fcn_rng_stream_data_available.__name__))
        elif fcn_rng_stream_data_available is None:
            self.cbkfcn_rng_stream_data_available = lambda: None
            self.lg.debug('{} Callback: Unregistering Stream Data Available function'.format(self.dev_name))
        else:
            self.lg.error('{} Callback: Provide fcn_stream_data_available as a function or None'.format(self.dev_name))


    def cbkreg_d2_input_capture_available(self, fcn_d2_input_capture_available):
        if callable(fcn_d2_input_capture_available):
            self.cbkfcn_d2_input_capture_available = fcn_d2_input_capture_available
            self.lg.debug('{} Callback: Registering D2 Input Capture Available function to {}'
                    .format(self.dev_name, fcn_d2_input_capture_available.__name__))
        elif fcn_d2_input_capture_available is None:
            self.cbkfcn_d2_input_capture_available = lambda: None
            self.lg.debug('{} Callback: Unregistering D2 Input Capture Available function'.format(self.dev_name))
        else:
            self.lg.error('{} Callback: Provide fcn_input_capture_available as a function or None'.format(self.dev_name))


    ## DEVICE

    def get_device_serial_number(self, timeout=GET_TIMEOUT_S):
        comm = 'DEVICE_SERIAL_NUMBER'
        if self.snd_rava_msg(comm):
            return self.get_queue_data(comm, timeout=timeout)


    def get_device_temperature(self, timeout=GET_TIMEOUT_S):
        comm = 'DEVICE_TEMPERATURE'
        if self.snd_rava_msg(comm):
            return self.get_queue_data(comm, timeout=timeout)


    def get_device_free_ram(self, timeout=GET_TIMEOUT_S):
        comm = 'DEVICE_FREE_RAM'
        if self.snd_rava_msg(comm):
            return self.get_queue_data(comm, timeout=timeout)


    def snd_device_reboot(self):
        comm = 'DEVICE_REBOOT'
        return self.snd_rava_msg(comm)


    ## EEPROM

    def snd_eeprom_reset_to_default(self):
        comm = 'EEPROM_RESET_TO_DEFAULT'
        return self.snd_rava_msg(comm)


    def snd_eeprom_device(self, temp_calib_slope, temp_calib_intercept):
        comm = 'EEPROM_DEVICE'
        rava_send = False
        data = [rava_send, temp_calib_slope, temp_calib_intercept]
        return self.snd_rava_msg(comm, data, 'BHh')


    def get_eeprom_device(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_DEVICE'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            data_vars = self.get_queue_data(comm, timeout=timeout)
            data_names = ['temp_calib_slope', 'temp_calib_intercept']
            return dict(zip(data_names, data_vars))


    def get_eeprom_firmware(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_FIRMWARE'
        if self.snd_rava_msg(comm):
            version_major, version_minor, version_patch, modules = self.get_queue_data(comm, timeout=timeout)

            version = '{}.{}.{}'.format(version_major, version_minor, version_patch)
            health_startup_enabled = modules & 1 << 0 != 0
            health_continuous_enabled = modules & 1 << 1 != 0
            led_enabled = modules & 1 << 2 != 0
            lamp_enabled = modules & 1 << 3 != 0
            peripherals_enabled = modules & 1 << 4 != 0
            serial1_enabled = modules & 1 << 5 != 0
            return {'version':version,
                    'health_startup_enabled':health_startup_enabled,
                    'health_continuous_enabled':health_continuous_enabled,
                    'led_enabled':led_enabled,
                    'lamp_enabled':lamp_enabled,
                    'peripherals_enabled':peripherals_enabled,
                    'serial1_enabled':serial1_enabled
                    }


    def snd_eeprom_pwm(self, freq_id, duty):
        if freq_id not in D_PWM_FREQ_INV:
            self.lg.error('{} EEPROM PWM: Unknown freq_id {}'.format(self.dev_name, freq_id))
            return False
        if duty == 0:
            self.lg.error('{} EEPROM PWM: Provide a duty > 0'.format(self.dev_name))
            return False

        comm = 'EEPROM_PWM'
        rava_send = False
        return self.snd_rava_msg(comm, [rava_send, freq_id, duty], 'BBB')


    def get_eeprom_pwm(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_PWM'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            freq_id, duty = self.get_queue_data(comm, timeout=timeout)
            return {'freq_id':freq_id, 'freq_str':D_PWM_FREQ_INV[freq_id], 'duty':duty}


    def snd_eeprom_rng(self, sampling_interval_us):
        if sampling_interval_us == 0:
            self.lg.error('{} EEPROM RNG: Provide a sampling_interval_us > 0'.format(self.dev_name))
            return False

        comm = 'EEPROM_RNG'
        rava_send = False
        return self.snd_rava_msg(comm, [rava_send, sampling_interval_us], 'BB')


    def get_eeprom_rng(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_RNG'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            sampling_interval_us = self.get_queue_data(comm, timeout=timeout)
            return {'sampling_interval_us':sampling_interval_us}


    def snd_eeprom_led(self, led_attached):
        comm = 'EEPROM_LED'
        rava_send = False
        return self.snd_rava_msg(comm, [rava_send, bool(led_attached)], 'BB')


    def get_eeprom_led(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_LED'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            led_attached = self.get_queue_data(comm, timeout=timeout)
            return {'led_attached':led_attached}


    def snd_eeprom_lamp(self, exp_dur_max_ms, exp_z_significant, exp_mag_smooth_n_trials):
        if exp_dur_max_ms < 60000:
            self.lg.error('{} EEPROM LAMP: Provide an exp_dur_max_ms >= 60000'.format(self.dev_name))
            return False
        if exp_z_significant < 1.0:
            self.lg.error('{} EEPROM LAMP: Provide an exp_z_significant >= 1.0'.format(self.dev_name))
            return False
        if exp_mag_smooth_n_trials == 0:
            self.lg.error('{} EEPROM LAMP: Provide an exp_mag_smooth_n_trials > 0'.format(self.dev_name))
            return False

        comm = 'EEPROM_LAMP'
        rava_send = False
        if self.snd_rava_msg(comm, [rava_send, exp_mag_smooth_n_trials, 8], 'BBB'):
            comm_extra = struct.pack('<Lf', exp_dur_max_ms, exp_z_significant)
            return self.write_serial(comm_extra)
        else:
            return False


    def get_eeprom_lamp(self, timeout=GET_TIMEOUT_S):
        comm = 'EEPROM_LAMP'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            data_vars = self.get_queue_data(comm, timeout=timeout)
            data_names = ['exp_dur_max_ms', 'exp_z_significant', 'exp_mag_smooth_n_trials']
            return dict(zip(data_names, data_vars))


    ## PWM

    def snd_pwm_setup(self, freq_id, duty):
        if freq_id not in D_PWM_FREQ_INV:
            self.lg.error('{} PWM: Unknown freq_id {}'.format(self.dev_name, freq_id))
            return False
        if duty == 0:
            self.lg.error('{} PWM: Provide a duty > 0'.format(self.dev_name))
            return False

        comm = 'PWM_SETUP'
        rava_send = False
        return self.snd_rava_msg(comm, [rava_send, freq_id, duty], 'BBB')


    def get_pwm_setup(self, timeout=GET_TIMEOUT_S):
        rava_send = True
        comm = 'PWM_SETUP'
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            freq_id, duty = self.get_queue_data(comm, timeout=timeout)
            return {'freq_id':freq_id, 'freq_str':D_PWM_FREQ_INV[freq_id], 'duty':duty}


    ## RNG

    def snd_rng_setup(self, sampling_interval_us):
        if sampling_interval_us == 0:
            self.lg.error('{} RNG: Provide a sampling_interval_us > 0'.format(self.dev_name))
            return False

        comm = 'RNG_SETUP'
        rava_send = False
        return self.snd_rava_msg(comm, [rava_send, sampling_interval_us], 'BB')


    def get_rng_setup(self, timeout=GET_TIMEOUT_S):
        comm = 'RNG_SETUP'
        rava_send = True
        if self.snd_rava_msg(comm, [rava_send], 'B'):
            sampling_interval_us = self.get_queue_data(comm, timeout=timeout)
            return {'sampling_interval_us':sampling_interval_us}


    def snd_rng_timing_debug_d1(self, on=True):
        comm = 'RNG_TIMING_DEBUG_D1'
        return self.snd_rava_msg(comm, [on], 'B')


    def get_rng_pulse_counts(self, n_counts, output_type='array', timeout=GET_TIMEOUT_S):
        if n_counts == 0:
            self.lg.error('{} RNG PC: Provide n_counts > 0'.format(self.dev_name))
            return None, None
        if n_counts >= 2**16:
            self.lg.error('{} RNG PC: Provide n_counts as a 16-bit integer'.format(self.dev_name))
            return None, None

        comm = 'RNG_PULSE_COUNTS'
        if self.snd_rava_msg(comm, [n_counts], 'H'):
            counts_bytes_a, counts_bytes_b = self.get_queue_data(comm, timeout=timeout)

            if (counts_bytes_a is None) or (counts_bytes_a is None):
                return None, None

            else:
                if output_type == 'list':
                    counts_a = bytes_to_list(counts_bytes_a, 'B')
                    counts_b = bytes_to_list(counts_bytes_b, 'B')
                    return counts_a, counts_b

                elif output_type == 'array':
                    counts_a = bytes_to_array(counts_bytes_a, np.uint8)
                    counts_b = bytes_to_array(counts_bytes_b, np.uint8)
                    return counts_a, counts_b

                else:
                    return counts_bytes_a, counts_bytes_b


    def get_rng_bits(self, bit_source_id, timeout=GET_TIMEOUT_S):
        if bit_source_id not in D_RNG_BIT_SRC_INV:
            self.lg.error('{} RNG Bits: Unknown bit_source_id {}'.format(self.dev_name, bit_source_id))
            return None

        comm = 'RNG_BITS'
        if self.snd_rava_msg(comm, [bit_source_id], 'B'):
            bit_source_id_recv, *bits = self.get_queue_data(comm, timeout=timeout)

            if bit_source_id == D_RNG_BIT_SRC['AB_RND']:
                bit_type_str = D_RNG_BIT_SRC_INV[bit_source_id_recv]
                return [bit_type_str, bits[0]]
            elif bit_source_id == D_RNG_BIT_SRC['AB']:
                return list(bits)
            else:
                return bits[0]


    def get_rng_bytes(self, n_bytes, postproc_id=D_RNG_POSTPROC['NONE'], request_id=0, output_type='array',
                      timeout=GET_TIMEOUT_S):
        if n_bytes == 0:
            self.lg.error('{} RNG Bytes: Provide n_bytes > 0'.format(self.dev_name))
            return None, None
        if n_bytes >= 2**16:
            self.lg.error('{} RNG Bytes: Provide n_bytes as a 16-bit integer'.format(self.dev_name))
            return None, None
        if postproc_id not in D_RNG_POSTPROC_INV:
            self.lg.error('{} RNG Bytes: Unknown postproc_id {}'.format(self.dev_name, postproc_id))
            return None, None

        comm = 'RNG_BYTES'
        if self.snd_rava_msg(comm, [n_bytes, postproc_id, request_id], 'HBB'):
            rng_bytes_a, rng_bytes_b = self.get_queue_data(comm, comm_ext_id=request_id, timeout=timeout)

            if (rng_bytes_a is None) or (rng_bytes_b is None):
                return None, None

            else:
                if output_type == 'list':
                    rng_a = bytes_to_list(rng_bytes_a, 'B')
                    rng_b = bytes_to_list(rng_bytes_b, 'B')
                    return rng_a, rng_b

                elif output_type == 'array':
                    rng_a = bytes_to_array(rng_bytes_a, np.uint8)
                    rng_b = bytes_to_array(rng_bytes_b, np.uint8)
                    return rng_a, rng_b

                else:
                    return rng_bytes_a, rng_bytes_b


    def get_rng_int8s(self, n_ints, int_delta, output_type='array', timeout=GET_TIMEOUT_S):
        # int_delta = int_max - int_min
        if n_ints == 0:
            self.lg.error('{} RNG Ints: Provide n_ints > 0'.format(self.dev_name))
            return None
        if n_ints >= 2**16:
            self.lg.error('{} RNG Ints: Provide n_ints as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta >= 2**8:
            self.lg.error('{} RNG Ints: Provide int_delta as a 8-bit integer'.format(self.dev_name))
            return None
        if int_delta == 0:
            self.lg.error('{} RNG Ints: Provide int_delta > 0'.format(self.dev_name))
            return None

        comm = 'RNG_INT8S'
        if self.snd_rava_msg(comm, [n_ints, int_delta], 'HB'):
            ints_bytes = self.get_queue_data(comm, timeout=timeout)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'B')
                else:
                    return bytes_to_array(ints_bytes, np.uint8)


    def get_rng_int16s(self, n_ints, int_delta, output_type='array', timeout=GET_TIMEOUT_S):
        # int_delta = int_max - int_min
        if n_ints == 0:
            self.lg.error('{} RNG Ints: Provide n_ints > 0'.format(self.dev_name))
            return None
        if n_ints >= 2**16:
            self.lg.error('{} RNG Ints: Provide n_ints as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta >= 2**16:
            self.lg.error('{} RNG Ints: Provide int_delta as a 16-bit integer'.format(self.dev_name))
            return None
        if int_delta == 0:
            self.lg.error('{} RNG Ints: Provide int_delta > 0'.format(self.dev_name))
            return None

        comm = 'RNG_INT16S'
        if self.snd_rava_msg(comm, [n_ints, int_delta], 'HH'):
            ints_bytes = self.get_queue_data(comm, timeout=timeout)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'H')
                else:
                    return bytes_to_array(ints_bytes, np.uint16)


    def get_rng_floats(self, n_floats, output_type='array', timeout=GET_TIMEOUT_S):
        if n_floats == 0:
            self.lg.error('{} RNG Floats: Provide n_floats > 0'.format(self.dev_name))
            return None
        if n_floats >= 2**16:
            self.lg.error('{} RNG Floats: Provide n_floats as a 16-bit integer'.format(self.dev_name))
            return None

        comm = 'RNG_FLOATS'
        if self.snd_rava_msg(comm, [n_floats], 'H'):
            ints_bytes = self.get_queue_data(comm, timeout=timeout)
            if ints_bytes is None:
                return None
            else:
                if output_type == 'list':
                    return bytes_to_list(ints_bytes, 'f')
                else:
                    return bytes_to_array(ints_bytes, np.float32)


    def snd_rng_byte_stream_start(self, n_bytes, stream_interval_ms, postproc_id=D_RNG_POSTPROC['NONE']):
        if n_bytes == 0:
            self.lg.error('{} RNG Stream: Provide n_bytes > 0'.format(self.dev_name))
            return None
        if n_bytes >= 2**16:
            self.lg.error('{} RNG Stream: Provide n_bytes as a 16-bit integer'.format(self.dev_name))
            return None
        if stream_interval_ms > RNG_BYTE_STREAM_MAX_INTERVAL_MS:
            self.lg.error('{} RNG Stream: Provide a stream_interval_ms <= {}ms.'
                     .format(self.dev_name, RNG_BYTE_STREAM_MAX_INTERVAL_MS))
            return None
        if postproc_id not in D_RNG_POSTPROC_INV:
            self.lg.error('{} RNG Stream: Unknown postproc_id {}'.format(self.dev_name, postproc_id))
            return None

        comm = 'RNG_STREAM_START'

        # Make queue empty
        while not self.serial_data[comm].empty():
            self.serial_data[comm].get_nowait()

        # Send command
        msg_success = self.snd_rava_msg(comm, [n_bytes, stream_interval_ms, postproc_id], 'HHB')
        if msg_success:
            self.rng_streaming = True

        return msg_success


    def snd_rng_byte_stream_stop(self):
        comm = 'RNG_STREAM_STOP'
        msg_success = self.snd_rava_msg(comm)

        if msg_success:
            self.rng_streaming = False

        return msg_success


    def get_rng_byte_stream_data(self, output_type='array', timeout=GET_TIMEOUT_S):
        if not self.rng_streaming:
            self.lg.warning('{} RNG Stream: Streaming is disabled'.format(self.dev_name))
            return None, None

        comm = 'RNG_STREAM_BYTES'
        rng_bytes_a, rng_bytes_b = self.get_queue_data(comm, timeout=timeout)

        if (rng_bytes_a is None) or (rng_bytes_b is None):
            return None, None

        else:
            if output_type == 'list':
                rng_a = bytes_to_list(rng_bytes_a, 'B')
                rng_b = bytes_to_list(rng_bytes_b, 'B')
                return rng_a, rng_b

            elif output_type == 'array':
                    rng_a = bytes_to_array(rng_bytes_a, np.uint8)
                    rng_b = bytes_to_array(rng_bytes_b, np.uint8)
                    return rng_a, rng_b

            else:
                return rng_bytes_a, rng_bytes_b


    def get_rng_byte_stream_status(self, timeout=GET_TIMEOUT_S):
        comm = 'RNG_STREAM_STATUS'
        if self.snd_rava_msg(comm):
            return self.get_queue_data(comm, timeout=timeout)


    ## HEALTH

    def snd_health_startup_run(self):
        comm = 'HEALTH_STARTUP_RUN'
        return self.snd_rava_msg(comm)


    def get_health_startup_results(self, timeout=GET_TIMEOUT_S):
        comm = 'HEALTH_STARTUP_RESULTS'
        if self.snd_rava_msg(comm):
            success, pc_avg_vars, pc_avg_diff_vars, bias_bit_vars, bias_byte_vars = self.get_queue_data(comm, timeout=timeout)
            data_vars = [pc_avg_vars, pc_avg_diff_vars, bias_bit_vars, bias_byte_vars]
            data_names = ['pulse_count', 'pulse_count_diff', 'bit_bias', 'byte_bias']
            return success, dict(zip(data_names, data_vars))


    def get_health_continuous_errors(self, timeout=GET_TIMEOUT_S):
        comm = 'HEALTH_CONTINUOUS_ERRORS'
        if self.snd_rava_msg(comm):
            data_vars = self.get_queue_data(comm, timeout=timeout)
            n_errors = sum(data_vars)
            data_names = ['repetitive_count_a', 'repetitive_count_b',
                          'adaptative_proportion_a', 'adaptative_proportion_b']
            return n_errors, dict(zip(data_names, data_vars))


    ## PERIPHERALS

    def snd_periph_digi_mode(self, periph_id, mode_id):
        if periph_id == 0 or periph_id > 5:
            self.lg.error('{} Periph: Provide a periph_id between 1 and 5'.format(self.dev_name))
            return None
        if mode_id not in D_PERIPH_MODES_INV:
            self.lg.error('{} Periph: Unknown mode_id {}'.format(self.dev_name, mode_id))
            return None

        comm = 'PERIPH_MODE'
        return self.snd_rava_msg(comm, [periph_id, mode_id], 'BB')


    def get_periph_digi_state(self, periph_id=1, timeout=GET_TIMEOUT_S):
        if periph_id == 0 or periph_id > 5:
            self.lg.error('{} Periph: Provide a periph_id between 1 and 5'.format(self.dev_name))
            return None

        comm = 'PERIPH_READ'
        if self.snd_rava_msg(comm, [periph_id], 'B'):
            digi_state, digi_mode = self.get_queue_data(comm, timeout=timeout)
            return digi_state, D_PERIPH_MODES_INV[digi_mode]


    def snd_periph_digi_state(self, periph_id, digi_state):
        if periph_id == 0 or periph_id > 5:
            self.lg.error('{} Periph: Provide a periph_id between 1 and 5'.format(self.dev_name))
            return None

        comm = 'PERIPH_WRITE'
        return self.snd_rava_msg(comm, [periph_id, digi_state], 'BB')


    def snd_periph_digi_pulse(self, periph_id=1, pulse_duration_us=100):
        if periph_id == 0 or periph_id > 5:
            self.lg.error('{} Periph: Provide a periph_id between 1 and 5'.format(self.dev_name))
            return None
        if pulse_duration_us == 0:
            self.lg.error('{} Periph: Provide a pulse_duration_us > 0'.format(self.dev_name))
            return None

        comm = 'PERIPH_PULSE'
        return self.snd_rava_msg(comm, [periph_id, pulse_duration_us], 'BH')


    def snd_periph_d1_trigger_input(self, on=True):
        comm = 'PERIPH_D1_TRIGGER_INPUT'
        return self.snd_rava_msg(comm, [on], 'B')


    def snd_periph_d1_comparator(self, neg_to_d5=False, on=True):
        comm = 'PERIPH_D1_COMPARATOR'
        return self.snd_rava_msg(comm, [on, neg_to_d5], 'BB')


    def snd_periph_d1_delay_us_test(self, delay_us):
        if delay_us == 0:
            self.lg.error('{} Periph: Provide a delay_us > 0'.format(self.dev_name))
            return None

        comm = 'PERIPH_D1_DELAY_US_TEST'
        return self.snd_rava_msg(comm, [delay_us], 'B')


    def snd_periph_d2_input_capture(self, on=True):
        comm = 'PERIPH_D2_TIMER3_INPUT_CAPTURE'

        # Make queue empty
        if on:
            while not self.serial_data[comm].empty():
                self.serial_data[comm].get_nowait()

        # Send command
        return self.snd_rava_msg(comm, [on], 'B')


    def get_periph_d2_input_capture(self, timeout=GET_TIMEOUT_S):
        comm = 'PERIPH_D2_TIMER3_INPUT_CAPTURE'
        return self.get_queue_data(comm, timeout=timeout)


    def snd_periph_d3_timer3_trigger_output(self, interval_ms=1, on=True):
        if interval_ms == 0 or interval_ms > RNG_BYTE_STREAM_MAX_INTERVAL_MS:
            self.lg.error('{} Periph: Provide a {} > interval_ms > 0'.format(self.dev_name, RNG_BYTE_STREAM_MAX_INTERVAL_MS))
            return None

        comm = 'PERIPH_D3_TIMER3_TRIGGER_OUTPUT'
        return self.snd_rava_msg(comm, [on, interval_ms], 'BH')


    def snd_periph_d3_timer3_pwm(self, freq_prescaler=1, top=2**8-1, duty=10,
                                 on=True):
        if freq_prescaler == 0 or freq_prescaler > 5:
            self.lg.error('{} Periph D3: Provide a freq_prescaler between 1 and 5'.format(self.dev_name))
            return None
        if top == 0:
            self.lg.error('{} Periph D3: Provide a top > 0'.format(self.dev_name))
            return None
        if duty == 0:
            self.lg.error('{} Periph D3: Provide a duty > 0'.format(self.dev_name))
            return None

        comm = 'PERIPH_D3_TIMER3_PWM'
        return self.snd_rava_msg(comm, [on, freq_prescaler, top, duty], 'BBHH')


    def snd_periph_d4_pin_change(self, on=True):
        comm = 'PERIPH_D4_PIN_CHANGE'
        return self.snd_rava_msg(comm, [on], 'B')


    def get_periph_d5_adc_read(self, ref_5v=1, clk_prescaler=0, oversampling_n_bits=0, on=True, timeout=GET_TIMEOUT_S):
        if clk_prescaler == 0 or clk_prescaler > 7:
            self.lg.error('{} Periph D5: Provide a clk_prescaler between 1 and 7'.format(self.dev_name))
            return None
        if oversampling_n_bits > 6:
            self.lg.error('{} Periph D5: Provide a oversampling_n_bits <= 6'.format(self.dev_name))
            return None

        comm = 'PERIPH_D5_ADC'
        if self.snd_rava_msg(comm, [on, ref_5v, clk_prescaler, oversampling_n_bits], 'BBBB'):
            if on:
                return self.get_queue_data(comm, timeout=timeout)


    ## INTERFACES

    def get_interface_ds18bs0(self, timeout=GET_TIMEOUT_S):
        comm = 'INTERFACE_DS18B20'
        if self.snd_rava_msg(comm):
            return self.get_queue_data(comm, timeout=timeout)


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
Definitions and variables used by the RAVA modules.
"""

FIRMWARE_MIN_VERSION = "1.0.0"

RAVA_USB_VID = 0x1209 # https://pid.codes
RAVA_USB_PID = 0x4884

COMM_MSG_START = b'$'
COMM_MSG_LEN = 8

SERIAL_LISTEN_LOOP_INTERVAL_S = 0.02 # 20 ms, 50 Hz readings
RNG_BYTE_STREAM_MAX_INTERVAL_MS = 4194
GET_TIMEOUT_S = 3.

PERIPH_PORTS = 5

LOG_FILL = 17 * ' '

D_DEV_COMM = {
    'DEVICE_SERIAL_NUMBER':1,
    'DEVICE_TEMPERATURE':2,
    'DEVICE_FREE_RAM':3,
    'DEVICE_REBOOT':4,
    'DEVICE_DEBUG':5,

    'EEPROM_RESET_TO_DEFAULT':10,
    'EEPROM_DEVICE':11,
    'EEPROM_FIRMWARE':12,
    'EEPROM_PWM':13,
    'EEPROM_RNG':14,
    'EEPROM_LED':15,
    'EEPROM_LAMP':16,

    'PWM_SETUP':30,

    'RNG_SETUP':40,
    'RNG_PULSE_COUNTS':41,
    'RNG_BITS':42,
    'RNG_BYTES':43,
    'RNG_TIMING_DEBUG_D1':44,

    'RNG_INT8S':50,
    'RNG_INT16S':51,
    'RNG_FLOATS':52,

    'RNG_STREAM_START':60,
    'RNG_STREAM_STOP':61,
    'RNG_STREAM_BYTES':62,
    'RNG_STREAM_STATUS':63,

    'HEALTH_STARTUP_RUN':70,
    'HEALTH_STARTUP_RESULTS':71,
    'HEALTH_CONTINUOUS_ERRORS':72,

    'LED_COLOR':80,
    'LED_COLOR_FADE':81,
    'LED_COLOR_OSCILLATE':82,
    'LED_INTENSITY':83,
    'LED_INTENSITY_FADE':84,
    'LED_FADE_STOP':85,
    'LED_STATUS':86,

    'LAMP_MODE':90,
    'LAMP_STATISTICS':91,
    'LAMP_DEBUG':92,

    'PERIPH_MODE':100,
    'PERIPH_READ':101,
    'PERIPH_WRITE':102,
    'PERIPH_PULSE':103,

    'PERIPH_D1_TRIGGER_INPUT':110,
    'PERIPH_D1_COMPARATOR':111,
    'PERIPH_D1_DELAY_US_TEST':112,
    'PERIPH_D2_TIMER3_INPUT_CAPTURE':113,
    'PERIPH_D3_TIMER3_TRIGGER_OUTPUT':114,
    'PERIPH_D3_TIMER3_PWM':115,
    'PERIPH_D4_PIN_CHANGE':116,
    'PERIPH_D5_ADC':117,

    'INTERFACE_DS18B20':130
    }
D_DEV_COMM_INV = {v: k for k, v in D_DEV_COMM.items()}

D_PWM_FREQ = {
    '30_KHZ':1,
    '40_KHZ':2,
    '50_KHZ':3,
    '60_KHZ':4,
    '75_KHZ':5
    }
D_PWM_FREQ_INV = {v: k for k, v in D_PWM_FREQ.items()}

D_RNG_BIT_SRC = {
    'AB':1,
    'A':2,
    'B':3,
    'AB_XOR':4,
    'AB_RND':5
}
D_RNG_BIT_SRC_INV = {v: k for k, v in D_RNG_BIT_SRC.items()}

D_RNG_POSTPROC = {
    'NONE':0,
    'XOR':1,
    'XOR_DICHTL':2,
    'VON_NEUMANN':3
    }
D_RNG_POSTPROC_INV = {v: k for k, v in D_RNG_POSTPROC.items()}

D_PERIPH_MODES = {
    'INPUT':0,
    'OUTPUT':1
}
D_PERIPH_MODES_INV = {v: k for k, v in D_PERIPH_MODES.items()}

D_LED_COLOR = {
    'RED':0,
    'ORANGE':16,
    'YELLOW':32,
    'GREEN':96,
    'CYAN':128,
    'BLUE':160,
    'PURPLE':192,
    'PINK':224
    }
D_LED_COLOR_INV = {v: k for k, v in D_LED_COLOR.items()}


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

__version__ = '1.2.0'

from .rava_defs import *
from .rava_rng import *
from .rava_rng_aio import RAVA_RNG_AIO
from .rava_rng_led import RAVA_RNG_LED

from . import tk
from . import acq


'''
This example illustrates the generation of random integers to a text file 
using the RAVA_ACQUISITION class.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Variables
OUT_FILE = True                         # File output
OUT_PATH = './'                         # Update with the desired output path
OUT_BINARY = False                      # Output as a text file
OUT_SEP = ','                           # Numbers separated by commas, CSV format
N_NUMS = 50000                          # 50k ints
N_CHUNK = 5000                          # Update progress every 5K ints
NUM_TYPE = int                          # Produce ints
NUM_MIN = 0                             # Int range [0, 999]
NUM_MAX = 999
THREADED = False                        # Blocking mode

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    print('No device found')
    exit()

# Create acquisition instance
rng_acq = rava.acq.RAVA_ACQUISITION()

# Register progress callback
rng_acq.cbkreg_progress(lambda progress: print('{:.0f}%'.format(progress)))

# Acquisition
outputs, progress, time_str = rng_acq.get_numbers(rng, N_NUMS, N_CHUNK, NUM_TYPE, NUM_MIN, NUM_MAX, 
                                                  OUT_FILE, OUT_PATH, OUT_BINARY, OUT_SEP, THREADED)

# Close RAVA device
rng.close()


'''
RAVA RNG usage example.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Find the serial number of the attached RAVA devices
rava_sns = rava.find_rava_sns()

# Create a RNG instance and connect to the first device
rng = rava.RAVA_RNG()
rng.connect(serial_number=rava_sns[0])

'''
The default PWM and RNG configuration parameters are stored in the EEPROM memory
and can be accessed with rng.get_eeprom_pwm() and rng.get_eeprom_rng(). If
desired, users can modify the default values using the respective snd_
functions. Additionally, it is possible to make non-permanent configuration
changes using the following commands:
'''

# Configure PWM
rng.snd_pwm_setup(freq_id=rava.D_PWM_FREQ['50_KHZ'], duty=20)

# Configure RNG
rng.snd_rng_setup(sampling_interval_us=10)

'''
Next, the generation of various random data types is demonstrated.
'''

# Measure 100 pulse counts
pc_a, pc_b = rng.get_rng_pulse_counts(n_counts=100)

# Generate a random bit XORing both channels
bit = rng.get_rng_bits(bit_source_id=rava.D_RNG_BIT_SRC['AB_XOR'])

# Generate 100 random bytes en each channel without post-processing
bytes_a, bytes_b = rng.get_rng_bytes(n_bytes=100,
                                     postproc_id=rava.D_RNG_POSTPROC['NONE'])

# Generate 100 8-bit integers between 0 and 99
ints8 = rng.get_rng_int8s(n_ints=100, int_delta=99)

# Generate 100 16-bit integers between 0 and 9999
ints16 = rng.get_rng_int16s(n_ints=100, int_delta=9999)

# Generate 100 32-bit floats ranging between 0 and 1
floats = rng.get_rng_floats(n_floats=100)

# Close RAVA device
rng.close()


'''
This example illustrates the generation of pulse counts to memory array using 
the RAVA_ACQUISITION class. The generation runs in threaded mode.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import rng_rava as rava

# Variables
OUT_FILE = False                        # Array output
N_PCS = 100000                           # 100K
N_CHUNK = 10000                         # Update progress every 10K
RNG_OUT = 'AB'                          # Generate 2 files, one for each RNG core
THREADED = True                         # Non-blocking mode

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    print('No device found')
    exit()

# Create acquisition instance
rng_acq = rava.acq.RAVA_ACQUISITION()

# Register progress callback
rng_acq.cbkreg_progress(lambda progress: print('{:.0f}%'.format(progress)))

# Define finished function
def finished(future):
    outputs, progress, time_str = future.result()

    # Print first 100 entries from each array
    print(outputs[0][:100])
    print(outputs[1][:100])
    
    # Close RAVA device
    rng.close()

# Acquisition
future = rng_acq.get_pulse_counts(rng, N_PCS, N_CHUNK, RNG_OUT, OUT_FILE, threaded=THREADED)
future.add_done_callback(finished)


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The SUBAPP_RAVA serves as the base class that every RAVA sub-application should
inherit from. It incorporates code that guarantees a unique instance for each
sub-application.
"""

import tkinter as tk


### RAVA_SUBAPP

class RAVA_SUBAPP(tk.Toplevel):

    cfg_default_str = ''

    def __init__(self, parent, name, win_title, win_geometry, win_resizable=False):
        # Initialize tk.Toplevel
        super().__init__(parent)
        self.parent = parent

        ## Initialize RAVA_SUBAPP
        self.name = name
        self.title(win_title)
        self.geometry('{}+{}+{}'.format(win_geometry, self.master.winfo_x() + self.master.winfo_width() + 10, self.master.winfo_y()))
        self.rowconfigure([0], weight=1)
        self.columnconfigure([0], weight=1)
        self.resizable(win_resizable, win_resizable)
        self.protocol('WM_DELETE_WINDOW', self.close)

        # Other instance already running? Destry this and keep the previous
        instances = [inst for inst in self.master.winfo_children() if isinstance(inst, type(self))]
        if len(instances) > 1:
            instances[-1].destroy()
            return False

        # Shortcuts
        self.lg = self.master.lg
        self.lgr = self.master.lgr
        self.cfg = self.master.cfg
        self.rng = self.master.rng

        # Debug
        self.lg.info('{}: Initializing'.format(self.name))

        # Key binds
        self.bind('<Control-Key-q>', lambda event=None: self.master.app_close())

        return True


    def close(self):
        # Destroy
        self.destroy()

        # Debug
        self.lg.info('{}: Closing'.format(self.name))


'''
This example showcases RNG byte stream.

This example code is in the public domain.
Author: Gabriel Guerrer
'''

import time
import numpy as np
import rng_rava as rava

N_BYTES = 12500
N_REPEAT = 3

# Find RAVA device and connect
rng = rava.RAVA_RNG()
dev_sns = rava.find_rava_sns()
if len(dev_sns):
    rng.connect(serial_number=dev_sns[0])
else:
    print('No device found')
    exit()

def rng_bytes_throughput(n_bytes, n_repeat, postproc_id):
    deltas = np.zeros(n_repeat)
    for i in range(n_repeat):
        t0 = time.perf_counter()
        rng.get_rng_bytes(n_bytes, postproc_id, timeout=None)
        t1 = time.perf_counter()
        deltas[i] = t1-t0

    delta_s = deltas.mean()
    freq = (n_bytes / delta_s) * 8 / 1000
    freq *= 2 # Considering both channels
    interv = (delta_s/ n_bytes) * 1e6
    print('Throughput: Produced {} x {} bytes in each channel'.format(n_bytes, n_repeat))
    print('  Freq Kbit/s = {:.3f}'.format(freq) )
    print('  Byte interval (us) = {:.3f}'.format(interv))

# Obtain the throughput for different post-processing options
for pp_str in ['NONE', 'XOR', 'XOR_DICHTL', 'VON_NEUMANN']:
    print('\n> PP {}'.format(pp_str))
    rng_bytes_throughput(n_bytes=N_BYTES, n_repeat=N_REPEAT, postproc_id=rava.D_RNG_POSTPROC[pp_str])

# Close device
rng.close()


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

from .rava_acquisition import RAVA_ACQUISITION, ACQ_PCS_RNG_OUT, ACQ_BYTES_RNG_OUT
from .acq_tools import get_ammount_prefix_number, get_ammount_prefix_str


"""
Copyright (c) 2024 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
Functions utilized by the Acquisition module.
"""

## GENERIC

def get_ammount_prefix_str(n):
    if n / (1000) < 1:
        ammount_str = '{:.0f}'.format(n)
    elif n / (1000000) < 1:
        ammount_str = '{:.0f}K'.format(n / 1000)
    elif n / (1000000000) < 1:
        ammount_str = '{:.0f}M'.format(n / 1000000)
    elif n / (1000000000000) < 1:
        ammount_str = '{:.0f}G'.format(n / 1000000000)
    elif n / (1000000000000000) < 1:
        ammount_str = '{:.0f}T'.format(n / 1000000000000)
    return ammount_str


def get_ammount_prefix_number(n, prefix):
    if 'K' in prefix.upper():
        return n * 1000
    elif 'M' in prefix.upper():
        return n * 1000000
    elif 'G' in prefix.upper():
        return n * 1000000000
    elif 'T' in prefix.upper():
        return n * 1000000000000


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
Configuration functionality for RAVA_APP using configparser's module.
"""

import os
import logging
import configparser


### RAVA_CFG

class RAVA_CFG:

    def __init__(self, cfg_filename, cfg_str):
        self.name = 'RAVA_CFG'
        self.lg = logging.getLogger('rava')        

        # Initialize parser
        self.cfg_filename = cfg_filename
        self.cfg_parser = self.parser_init(cfg_filename, cfg_str)


    # Low level functionality

    def parser_init(self, cfg_filename, cfg_str):
        # Default cfg
        cfg_default = configparser.ConfigParser()
        cfg_default.read_string(cfg_str)

        # Config file already exists?
        if os.path.isfile(cfg_filename):

            # Read cfg info
            cfg_parser = configparser.ConfigParser()
            cfg_parser.read(cfg_filename)

            # The read cfg contains all the expected entries?
            for key in cfg_default:
                if key not in cfg_parser or set(cfg_parser[key].keys()) != set(cfg_default[key].keys()):
                    os.remove(cfg_filename)
                    self.lg.error('{} Config: The provided config file is incomplete. Please restart the application.'.format(self.name))
                    exit()

        # Config file inexists, create it with default values
        else:
            self.lg.info('{} Config: Creating file with default content'.format(self.name))
            cfg_parser = cfg_default
            
            # Save
            with open(cfg_filename, 'w') as cfg_file:
                cfg_parser.write(cfg_file)

        return cfg_parser


    # High level functionality

    def read(self, section, option, type=str):
        # Return config option in the appropriate type
        if type is int:
            data = self.cfg_parser.getint(section, option)
        elif type is float:
            data = self.cfg_parser.getfloat(section, option)
        elif type is bool:
            data = self.cfg_parser.getboolean(section, option)
        else:
            data = self.cfg_parser.get(section, option)

        return data


    def write(self, section, option, value, save=True):
        # Write config option
        if type(value) is not str:
            value = str(value)
        self.cfg_parser.set(section, option, value)
        
        # Save?
        if save:
            with open(self.cfg_filename, 'w') as cfg_file:
                self.cfg_parser.write(cfg_file)


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
This code evokes the RAVA's Tk Control Panel application.

To execute it, enter the following command: python3 -m rng_rava.tk.ctrlp
"""

import rng_rava as rava 

TITLE = 'RAVA Control Panel'

SUBAPPS = [rava.tk.ctrlp.rava_subapp_control_panel]


def main():
    # RAVA main app
    tkapp = rava.tk.RAVA_APP(title=TITLE, subapp_dicts=SUBAPPS, rava_class=rava.RAVA_RNG_LED, cfg_log_name='rava_ctrlp')

    # Enter Tk loop
    tkapp.mainloop()


if __name__ == '__main__':
    main()


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The RAVA_APP is a Tk application designed to serve as a foundational framework
for various Graphical User Interface (GUI) sub-applications utilizing the RAVA
device.

The RAVA_APP instance denoted as the main application, lists the available
RAVA devices and handles the connection to the selected circuit, detecting and
responding to eventual disconnections.

The device's communication data and the program's state can be visualized in a
dedicated log window, a valuable resource for debugging the applications and
reporting any encountered errors.

Additionally, the main application features a status bar for vital information
display, and configuration file utilities for the permanent storage of user
preferences.

Sub-applications are tk.Toplevel windows that integrate into the main
application. Derived from the RAVA_SUBAPP class, they become part of the main
application when included in the subapp_dicts argument. Once integrated, these
sub-applications are easily accessible through the main application's menu and
central buttons.
"""

import sys
import os
import os.path
from functools import partial
import logging

import tkinter as tk
from tkinter import ttk
import tkinter.messagebox as tkm
import tkinter.filedialog as tkfd

from rng_rava.tk import RAVA_CFG, RAVA_SUBAPP
from rng_rava import find_rava_sns, RAVA_RNG, RAVA_RNG_LED


### Parameters

PAD = 10
HOME_RAVA_PATH = os.path.join(os.path.expanduser('~'), '.rava/')
SUBAPP_DICT_KEYS = ['class', 'menu_title', 'show_button', 'use_rng']

LOG_NAME_RAVA_DRV = 'rava'
LOG_NAME_RAVA_APP = 'rava_app'
LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR']


### RAVA_APP

class RAVA_APP(tk.Tk):

    cfg_default_str = '''
    [RAVA_LOG]
    level_rava_driver = INFO
    level_rava_app = INFO
    show_on_startup = True
    update_interval_ms = 500

    [RAVA_APP]
    connect_on_startup = True
    status_duration_ms = 3000
    '''

    def __init__(self, title, geometry='480x300', subapp_dicts=[], rava_class=RAVA_RNG, cfg_log_name='rava'):
        # Evaluate parameters
        if not self.subapp_evaluate(subapp_dicts):
            print('RAVA_APP Error: subapp_dicts must be suplied as a list of dictionaries containing the following keys: {}'
                  .format(SUBAPP_DICT_KEYS))
            exit()
        self.subapp_dicts = subapp_dicts
        self.subapp_buttons = {}

        ### Initialize RAVA_APP
        super().__init__()
        self.name = 'RAVA_APP'
        self.title(title)
        self.geometry(geometry)
        self.rowconfigure([0,1], weight=1)
        self.columnconfigure([0], weight=1)
        self.protocol('WM_DELETE_WINDOW', self.app_close)
        self.resizable(False, False) # No resize
        self.option_add('*tearOff', tk.FALSE) # Diable menu tear-off
        self.update() # Update geometry       
        
        # Files
        if not os.path.exists(HOME_RAVA_PATH):
            os.mkdir(HOME_RAVA_PATH)
    
        self.filename_logo = os.path.join(os.path.dirname(__file__), 'rava.png')
        self.filename_cfg = os.path.join(HOME_RAVA_PATH, '{}.cfg'.format(cfg_log_name))
        self.filename_log = os.path.join(HOME_RAVA_PATH, '{}.log'.format(cfg_log_name))

        # Logger setup        
        self.lgr, self.lg = self.log_setup()

        # Config setup
        cfg_subapp_str = [subapp_dict['class'].cfg_default_str for subapp_dict in self.subapp_dicts if subapp_dict['class'].cfg_default_str]
        cfg_str = '\n'.join([self.cfg_default_str] + cfg_subapp_str)
        
        self.cfg = RAVA_CFG(self.filename_cfg, cfg_str)

        # Catch Tk exeptions
        self.report_callback_exception = self.app_handle_tk_exception

        # Debug
        self.lg.info('{}: Initializing'.format(self.name))

        ## Widgets

        # Log Window
        self.window_log = WIN_LOG(self)

        # Menu
        self.widgets_menu()

        # RAVA Connection
        self.frm_rava_conn = ttk.Frame(self, padding=(PAD, PAD))
        self.frm_rava_conn.grid(row=0, column=0, sticky='nsew')
        self.widgets_rava_connection()

        # Center
        self.frm_center =ttk.Frame(self, padding=(PAD, PAD))
        self.frm_center.grid(row=1, column=0, sticky='nsew')
        self.widgets_center()

        # Status Footer
        self.frm_status = ttk.Frame(self, padding=(4, 4))
        self.frm_status.grid(row=2, column=0, sticky='ew')
        self.widgets_status()
        
        self.task_status = None

        ## Key Binds
        self.bind('<Control-Key-l>', lambda event=None: self.window_log.show())
        self.bind('<Control-Key-s>', lambda event=None: self.rava_scan())
        self.bind('<Control-Key-q>', lambda event=None: self.app_close())

        ### Start

        # Update widgets enabled state
        self.widgets_state(rava_connected=False)

        # Initialize RAVA Device
        if not rava_class in [RAVA_RNG, RAVA_RNG_LED]:
            self.lg.error('{}: Please provide rava_class as RAVA_RNG or RAVA_RNG_LED'.format(self.name))
            return
        
        self.rng = rava_class()

        # Register disconnect callback
        self.rng.cbkreg_device_close(self.rava_disconnect)

        # Scan for RAVA devices and connect if one found
        if self.cfg.read('RAVA_APP', 'connect_on_startup', bool):
            self.rava_scan()


    ## LOGGING

    def log_setup(self):
        # Logger Formatter / Handler
        lg_fmt = logging.Formatter(fmt='%(asctime)s %(levelname)-7s %(message)s', datefmt='%H:%M:%S')
        lg_sh = logging.StreamHandler()
        lg_sh.setFormatter(lg_fmt)
        lg_fh = logging.FileHandler(self.filename_log, mode='w')
        lg_fh.setFormatter(lg_fmt)

        # Logger : RAVA Driver
        lgr = logging.getLogger(LOG_NAME_RAVA_DRV)
        if not lgr.hasHandlers():
            lgr.addHandler(lg_sh)
            lgr.addHandler(lg_fh)
            lgr.setLevel(logging.INFO)

        # Logger: RAVA APP
        lg = logging.getLogger(LOG_NAME_RAVA_APP)
        if not lg.hasHandlers():
            lg.addHandler(lg_sh)
            lg.addHandler(lg_fh)
            lg.setLevel(logging.INFO)

        # Unhandled Exceptions 
        def log_unhandled_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback) # Call default excepthook
                return
            self.lg.critical('Unhandled Exception', exc_info=(exc_type, exc_value, exc_traceback))

        sys.excepthook = log_unhandled_exception

        return lgr, lg


    ## APP

    def app_handle_tk_exception(self, exc_type, exc_value, exc_traceback):
        self.lg.critical('{}: TK Exception'.format(self.name), exc_info=(exc_type, exc_value, exc_traceback))


    def app_close(self):
        # Debug
        self.lg.info('{}: Closing'.format(self.name))

        # Unregister RAVA close callback
        self.rng.cbkreg_device_close(None)

        # Close rava device
        self.rng.close()

        # Close Log Window
        self.window_log.close()

        # Close tkinter
        self.destroy()
        self.quit()


    def subapp_evaluate(self, subapp_dicts):
        # Test if subapp_dicts is a list of dictionaries containing SUBAPP_DICT_KEYS keys
        for subapp_dict in subapp_dicts:
            if not (isinstance(subapp_dict, dict) and set(subapp_dict.keys()) == set(SUBAPP_DICT_KEYS)):                
                return False
        
        # Test values
        for subapp_dict in subapp_dicts:

            # Class inherits from RAVA_SUBAPP
            if not issubclass(subapp_dict['class'], RAVA_SUBAPP):
                return False

            # Default values            
            if subapp_dict['show_button'] is None:
                subapp_dict['cfg_log_name'] = True
            if subapp_dict['use_rng'] is None:
                subapp_dict['cfg_log_name'] = True
            
        return True


    ## WIDGETS

    def widgets_menu(self):
        self.menu = tk.Menu(self)
        self['menu'] = self.menu

        ## File
        self.menu_file = tk.Menu(self.menu)
        self.menu.add_cascade(menu=self.menu_file, label='File')

        # Log
        self.menu_file.add_command(label='Show Log', command=self.window_log.show, underline=5)
        self.menu_file.add_separator()

        # Quit
        self.menu_file.add_command(label='Quit', command=self.app_close, underline=0)

        ## Sub-Apps
        self.menu_subapps = tk.Menu(self.menu)
        self.menu.add_cascade(menu=self.menu_subapps, label='Sub-Appps')

        # Create a menu entry for all subapps
        for subapp_dict in self.subapp_dicts:
            self.menu_subapps.add_command(label=subapp_dict['menu_title'], command=partial(subapp_dict['class'], self))


    def widgets_rava_connection(self):
        self.frm_rava_conn.columnconfigure([3], weight=1)
        self.frm_rava_conn.rowconfigure([0], weight=1)

        self.lb_rava_sn = ttk.Label(self.frm_rava_conn, text='SN')
        self.lb_rava_sn.grid(row=0, column=0, sticky='w', padx=(PAD, 0))

        self.cbb_rava_sns = ttk.Combobox(self.frm_rava_conn, width=12)
        self.cbb_rava_sns.grid(row=0, column=1, sticky='w', padx=(PAD,0))
        self.cbb_rava_sns.state(['readonly'])
        self.cbb_rava_sns.bind('<<ComboboxSelected>>', self.rava_connect)

        self.bt_rava_scan = ttk.Button(self.frm_rava_conn, text='Scan', width=7, command=self.rava_scan, underline=0)
        self.bt_rava_scan.grid(row=0, column=2, sticky='w', padx=(PAD,0))

        self.lb_rava_conn = ttk.Label(self.frm_rava_conn, anchor='c', text='Disconnected')
        self.lb_rava_conn.grid(row=0, column=3, sticky='we', padx=(2*PAD,0))


    def widgets_center(self):
        subapp_dicts_show = [subapp_dict for subapp_dict in self.subapp_dicts if subapp_dict['show_button']]
        n_apps = len(subapp_dicts_show)
        rows = list(range(n_apps)) if n_apps else [0]
        logo_rowspan = n_apps if n_apps else 1

        self.frm_center.columnconfigure([0,1], weight=1)
        self.frm_center.rowconfigure(rows, weight=1)

        # RAVA Logo
        self.img_rava = tk.PhotoImage(file=self.filename_logo)
        self.lb_img_rava = ttk.Label(self.frm_center, image=self.img_rava)
        self.lb_img_rava.grid(row=0, column=0, rowspan=logo_rowspan)

        # Buttons
        # Create buttons for all subapp_dicts with the option "show_button = True"
        for i, subapp_dict in enumerate(subapp_dicts_show):
            bt = ttk.Button(self.frm_center, width=20, text=subapp_dict['menu_title'], command=partial(subapp_dict['class'], self))
            bt.grid(row=i, column=1)
            self.subapp_buttons[subapp_dict['menu_title']] = bt


    def widgets_status(self):
        self.frm_status.columnconfigure([0], weight=1)
        self.frm_status.rowconfigure([0], weight=1)

        self.lb_status = ttk.Label(self.frm_status, text='', relief=tk.RIDGE, anchor='e')
        self.lb_status.grid(row=0, column=0, sticky='ew')


    def widgets_state(self, rava_connected):
        if rava_connected:
            menu_state = tk.NORMAL
            bt_state = ['!disabled']
        else:
            menu_state = tk.DISABLED
            bt_state = ['disabled']

        for subapp_dict in self.subapp_dicts:
            if subapp_dict['use_rng']:
                # Menu
                self.menu_subapps.entryconfigure(subapp_dict['menu_title'], state=menu_state)

                # Buttons
                if subapp_dict['show_button']:
                    self.subapp_buttons[subapp_dict['menu_title']].state(bt_state)


    def rava_scan(self):
        # Find serial numbers and actualize combobox values
        sns = find_rava_sns()
        self.cbb_rava_sns['values'] = sns

        # Status bar
        self.status_set('{} devices found'.format(len(sns)))

        # No one found
        if len(sns) == 0:
            self.cbb_rava_sns.set('')

        # If only one found, connect
        elif len(sns) == 1:
            self.cbb_rava_sns.set(sns[0])
            self.rava_connect()


    def rava_connect(self, tk_event=None):
        # Remove blue selection
        self.selection_clear()

        # Get serial number
        sn = self.cbb_rava_sns.get()
        if not sn:
            return

        # Previously connected ...
        if self.rng.connected():
            # ... to the same sn? Leave
            if self.rng.dev_serial_number == sn:
                return
            # ... to other sn? Disconnect
            else:
                self.rng.close()

        # Debug
        self.lg.info('{}: Connecting to RAVA device...'.format(self.name))

        # Connect
        if (self.rng.connect(sn)):
            # SN Combobox
            self.cbb_rava_sns.set(sn)

            # Connection label
            self.lb_rava_conn['text'] = 'Connected to {}'.format(self.rng.dev_usb_port)

            # Status bar
            self.status_set('Connected to RAVA device')

            # Update widgets state
            self.widgets_state(True)

        else:
            # SN Combobox
            self.cbb_rava_sns.set('')

            # Connection label
            self.lb_rava_conn['text'] = 'Disconnected'

            # Status bar
            self.status_set('Error connecting to RAVA device')

            # Update widgets state
            self.widgets_state(False)

            # Startup Health tests failed?
            if self.rng.health_startup_enabled and self.rng.health_startup_success == False:
                tkm.showerror(parent=self, message='Startup Health Tests Failed!')
                self.rng.close()


    def rava_disconnect(self):
        # Debug
        self.lg.info('{}: Lost connection to RAVA device'.format(self.name))

        # Update SN Combobox
        self.cbb_rava_sns['values'] = find_rava_sns()
        self.cbb_rava_sns.set('')

        # Connection label
        self.lb_rava_conn['text'] = 'Disconnected'

        # Status bar
        self.status_set('RAVA device disconnected')

        # Update widgets state
        self.widgets_state(False)

        # Close windows that use rng
        for subapp_dict in self.subapp_dicts:
            if subapp_dict['use_rng']:
                for inst in self.winfo_children():
                    if isinstance(inst, subapp_dict['class']):
                        inst.close()


    def status_set(self, str):
        # Cancel last task
        if self.task_status is not None:
            self.after_cancel(self.task_status)

        # Update
        self.lb_status['text'] = str + '  '

        # Erase after status_duration_ms
        if str:
            status_duration_ms = self.cfg.read('RAVA_APP', 'status_duration_ms', int)
            self.task_status = self.after(status_duration_ms, self.status_set, '')


### WIN_LOG

class WIN_LOG(tk.Toplevel):

    def __init__(self, parent):
        ## Initialize WIN_LOG
        super().__init__(parent)
        self.name = 'WIN_LOG'
        self.title('RAVA Log')
        self.geometry('750x500+{}+{}'.format(self.master.winfo_x(), self.master.winfo_y() + self.master.winfo_height() + 40))
        self.rowconfigure([0], weight=1)
        self.columnconfigure([0], weight=1)
        self.protocol('WM_DELETE_WINDOW', self.hide) # One instance

        # Logging / cfg
        self.lg = self.master.lg
        self.cfg = self.master.cfg

        # Widgets
        self.widgets()

        # KEY BINDS
        self.bind('<Control-Key-q>', lambda event=None: self.master.app_close())

        ## Start
        show_on_startup = self.cfg.read('RAVA_LOG', 'show_on_startup', bool)
        if not show_on_startup:
            self.hide()

        # Read cfg and update widgets
        self.cfg_read_update_widgets()

        # Set the log level according to the widgets values
        self.log_level_rava_driver(write=False)
        self.log_level_rava_app(write=False)

        # Open log file and keep updating it
        self.task_log_update = None
        self.log_file = open(self.master.filename_log, 'r')
        self.log_update()


    def widgets(self):
        self.frm_log = ttk.Frame(self, padding=(PAD, PAD))
        self.frm_log.grid(row=0, column=0, sticky='nsew')
        self.frm_log.columnconfigure([0], weight=1)
        self.frm_log.rowconfigure([0], weight=1)

        self.txt_log = tk.Text(self.frm_log, state='disabled')
        self.txt_log.grid(row=0, column=0, sticky='nsew')

        self.scroll_y = ttk.Scrollbar(self.frm_log, orient=tk.VERTICAL, command=self.txt_log.yview)
        self.scroll_y.grid(row=0, column=1, sticky='ns')
        self.txt_log['yscrollcommand'] = self.scroll_y.set

        self.frm_log_actions = ttk.Frame(self, padding=(PAD, 0, PAD, PAD))
        self.frm_log_actions.grid(row=1, column=0, columnspan=2, sticky='nsew')
        self.frm_log_actions.columnconfigure([0,1,2,3,4,5,6], weight=1)
        self.frm_log_actions.rowconfigure([0], weight=1)

        self.lb_rava_driver_level = ttk.Label(self.frm_log_actions, text='RAVA driver')
        self.lb_rava_driver_level.grid(row=0, column=0)

        self.cbb_rava_driver_level = ttk.Combobox(self.frm_log_actions, width=8)
        self.cbb_rava_driver_level.grid(row=0, column=1, sticky='w')
        self.cbb_rava_driver_level['values'] = LOG_LEVELS
        self.cbb_rava_driver_level.state(['readonly'])
        self.cbb_rava_driver_level.bind('<<ComboboxSelected>>', self.log_level_rava_driver)

        self.lb_rava_app_level = ttk.Label(self.frm_log_actions, text='RAVA app')
        self.lb_rava_app_level.grid(row=0, column=2)

        self.cbb_rava_app_level = ttk.Combobox(self.frm_log_actions, width=8)
        self.cbb_rava_app_level.grid(row=0, column=3, sticky='w')
        self.cbb_rava_app_level['values'] = LOG_LEVELS
        self.cbb_rava_app_level.state(['readonly'])
        self.cbb_rava_app_level.bind('<<ComboboxSelected>>', self.log_level_rava_app)

        self.var_show_startup = tk.BooleanVar(value=False)
        self.cb_show_startup = ttk.Checkbutton(self.frm_log_actions, text='Show on startup',
                                               variable=self.var_show_startup, command=self.log_win_show_startup)
        self.cb_show_startup.grid(row=0, column=4)

        self.bt_copy = ttk.Button(self.frm_log_actions, text='Copy to Clipboard', command=self.log_copy)
        self.bt_copy.grid(row=0, column=5)

        self.bt_save = ttk.Button(self.frm_log_actions, text='Save File', command=self.log_save)
        self.bt_save.grid(row=0, column=6)


    def show(self):
        self.geometry('750x500+{}+{}'.format(self.master.winfo_x(), self.master.winfo_y() + self.master.winfo_height() + 10))
        self.wm_deiconify()
        self.bt_save.focus()


    def hide(self):
        self.withdraw()


    def close(self):
        # Cancel scheduled log update
        self.after_cancel(self.task_log_update)
        self.destroy()


    def log_update(self):
        # Update log content
        new_data = self.log_file.read()
        if new_data:
            self.txt_log.configure(state='normal')
            self.txt_log.insert('end', new_data)
            self.txt_log.see('end')
            self.txt_log.configure(state='disabled')

        # Schedule new update
        update_interval_ms = self.cfg.read('RAVA_LOG', 'update_interval_ms', int)
        self.task_log_update = self.after(update_interval_ms, self.log_update)


    def cfg_read_update_widgets(self):
        # level_rava_driver
        level_rava_driver_str = self.cfg.read('RAVA_LOG', 'level_rava_driver')

        if level_rava_driver_str in LOG_LEVELS:
            self.cbb_rava_driver_level.set(level_rava_driver_str)
        else:
            self.lg.warning('{}: Unknown level_rava_driver={}. Provide one of those:{}'.format(self.name, level_rava_driver_str, LOG_LEVELS))
            self.cbb_rava_driver_level.set('INFO')

        # level_rava_app
        level_rava_app_str = self.cfg.read('RAVA_LOG', 'level_rava_app')

        if level_rava_app_str in LOG_LEVELS:
            self.cbb_rava_app_level.set(level_rava_app_str)
        else:
            self.lg.warning('{}: Unknown level_rava_app={}. Provide one of those:{}'.format(self.name, level_rava_app_str, LOG_LEVELS))
            self.cbb_rava_app_level.set('INFO')

        # show_on_startup
        show_on_startup = self.cfg.read('RAVA_LOG', 'show_on_startup', bool)
        self.var_show_startup.set(show_on_startup)


    def log_level_rava_driver(self, tk_event=None, write=True):
        # Update log level
        level_rava_driver_str = self.cbb_rava_driver_level.get()
        level_rava_driver = getattr(logging, level_rava_driver_str)
        self.master.lgr.setLevel(level_rava_driver)

        # Update cfg
        if write:
            self.cfg.write('RAVA_LOG', 'level_rava_driver', level_rava_driver_str)


    def log_level_rava_app(self, tk_event=None, write=True):
        # Update log level
        level_rava_app_str = self.cbb_rava_app_level.get()
        level_rava_app = getattr(logging, level_rava_app_str)
        self.master.lg.setLevel(level_rava_app)

        # Update cfg
        if write:
            self.cfg.write('RAVA_LOG', 'level_rava_app', level_rava_app_str)


    def log_win_show_startup(self, tk_event=None, write=True):
        # Update cfg
        if write:
            self.cfg.write('RAVA_LOG', 'show_on_startup', self.var_show_startup.get())


    def log_copy(self):
        # Copy log content to clipboard
        self.master.clipboard_clear()
        self.master.clipboard_append(self.txt_log.get('1.0', 'end'))


    def log_save(self):
        # Get filename
        filename = tkfd.asksaveasfilename(filetypes=[('LOG','.log')])

        # Save
        if filename:
            with open(filename, 'w') as file:
                file.write(self.txt_log.get('1.0', 'end'))

            # Status
            self.master.status_set('Log saved')


"""
Copyright (c) 2024 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
The RAVA_ACQUISITION class implements the generation of extensive datasets
comprising random bytes, numbers or pulse counts extracted from a RAVA device.
Generation can occur in either threaded or blocking mode, and the output
can be presented as a memory array or a disk file.

For the get_pulse_counts() and get_bytes() functions, when out_file=True, they 
generate binary files. In contrast, the get_numbers() function provides the 
option to produce a text file by setting the parameter out_file_binary=False. 
In the text output, entries are separated by the out_separator character. 
When get_numbers() has the parameter out_file_binary=True, the files are 
saved using the numpy npy file format.

The get_numbers() function generates one file or array, which alternates entries 
from both RNG cores. The get_pulse_counts() and get_bytes() functions accept a 
rng_out parameter, which can be:
* 'A': 1 output file, RNG core A only
* 'B': 1 output file, RNG core B only
* 'AB': 2 output files, RNG cores A and B independent data

For get_bytes(), the parameter rng_out also accepts:
* 'AB_XOR': 1 output file, obtained by XORing RNG cores A and B output
* 'AB_ALT': 1 output file, with alternating bytes from RNG cores A and B. 
   Requires half of the time to produce n_bytes

A progress callback is established by registering a function through 
cbkreg_progress(). This function is invoked every time a data chunk of size 
n_chunk is generated, providing information about the progress percentage.
"""

import os.path
from glob import glob
import time
import logging
import itertools
import concurrent.futures
import threading
import datetime

import numpy as np

from rng_rava.acq.acq_tools import get_ammount_prefix_str
from rng_rava.rava_defs import D_RNG_POSTPROC, LOG_FILL


### VARS

ACQ_PCS_RNG_OUT = ['A', 'B', 'AB']
ACQ_BYTES_RNG_OUT = ['A', 'B', 'AB', 'AB_XOR', 'AB_ALT']


### RAVA_ACQUISITION

class RAVA_ACQUISITION:

    def __init__(self, delete_incomplete=True):
        # Set vars
        self.name = 'RAVA_ACQUISITION'
        self.delete_incomplete = delete_incomplete

        # Logging
        self.lg = logging.getLogger('rava')

        # Threading
        self.acquiring = threading.Event()
        self.thread_executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)

        # Callback functions
        self.cbkfcn_progress = lambda progress: None


    ## CALLBACKS

    def cbkreg_progress(self, fcn_acquire_progress):
        if callable(fcn_acquire_progress):
            self.cbkfcn_progress = fcn_acquire_progress
            self.lg.debug('{} Callback: Registering Progress function to {}'
                          .format(self.name, fcn_acquire_progress.__name__))
        elif fcn_acquire_progress is None:
            self.cbkfcn_progress = lambda progress: None
            self.lg.debug('{} Callback: Unregistering Progress function'.format(self.name))
        else:
            self.lg.error('{} Callback: Provide fcn_acquire_progress as a function or None'.format(self.name))


    ## GENERAL

    def stop(self):
        self.lg.info('{}: Stopping acquisition'.format(self.name))
        self.acquiring.clear()


    ## PULSE COUNTS

    def get_filenames_pulse_counts(self, rng, n_pcs, rng_out, out_path):
        # RNG serial number
        sn = rng.dev_serial_number

        # Filenames base
        ammount_str = get_ammount_prefix_str(n_pcs)
        filename_a = '{}_PCS_{}_A__'.format(sn, ammount_str)
        filename_b = '{}_PCS_{}_B__'.format(sn, ammount_str)

        # Filenames count
        n_filename_a = len(glob(os.path.join(out_path, filename_a + '*.bin')))
        n_filename_b = len(glob(os.path.join(out_path, filename_b + '*.bin')))
        n_filename_ab = max(n_filename_a, n_filename_b)

        # Result Filename: count corrected + path
        if rng_out == 'A':
            filename_a = os.path.join(out_path, '{}{}.bin'.format(filename_a, n_filename_a + 1))
            return filename_a, None

        elif rng_out == 'B':
            filename_b = os.path.join(out_path, '{}{}.bin'.format(filename_b, n_filename_b + 1))
            return None, filename_b

        elif rng_out == 'AB':
            filename_a = os.path.join(out_path, '{}{}.bin'.format(filename_a, n_filename_ab + 1))
            filename_b = os.path.join(out_path, '{}{}.bin'.format(filename_b, n_filename_ab + 1))
            return filename_a, filename_b


    def get_pulse_counts(self, rng, n_pcs, n_chunk=10000, rng_out='AB', out_file=False, out_path='', threaded=False):
        # Test function inputs
        if not rng.connected():
            self.lg.error('{} Pulse Counts: Trying to acquire data from a disconnected RNG'.format(self.name))
            return [], 0, ''

        if self.acquiring.is_set():
            self.lg.error('{} Pulse Counts: RNG is already acquiring'.format(self.name))
            return [], 0, ''

        if out_file and not os.path.isdir(out_path):
            self.lg.error('{} Pulse Counts: The provided out_path doesn\'t exists'.format(self.name))
            return [], 0, ''

        if n_pcs <= 0:
            self.lg.error('{} Pulse Counts: Provide n_pcs > 0'.format(self.name))
            return [], 0, ''

        if n_chunk <= 0:
            self.lg.error('{} Pulse Counts: Provide n_chunk > 0'.format(self.name))
            return [], 0, ''

        if rng_out not in ACQ_PCS_RNG_OUT:
            self.lg.error('{} Pulse Counts: Unknown rng_out. Please choose from {}'.format(self.name, ACQ_PCS_RNG_OUT))
            return [], 0, ''

        # Start generation
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Pulse Counts: Starting {} acquisition with device {}'
                     .format(self.name, out_type, rng.dev_serial_number))

        args = (rng, n_pcs, n_chunk, rng_out, out_file, out_path)

        if threaded:
            result_future = self.thread_executor.submit(self.loop_pulse_counts, *args)
            return result_future
        else:
            result = self.loop_pulse_counts(*args)
            return result


    def loop_pulse_counts(self, rng, n_pcs, n_chunk, rng_out, out_file, out_path):
        # Set variables
        time_start = time.perf_counter()
        n_acquired = 0
        n_remaining = n_pcs
        progress = 0.
        self.acquiring.set()

        # Output
        out_path = '' if not out_file else out_path
        filename_a, filename_b = self.get_filenames_pulse_counts(rng, n_pcs, rng_out, out_path)

        if out_file:
            f_a = open(filename_a, mode='bw') if filename_a else None
            f_b = open(filename_b, mode='bw') if filename_b else None
            output_type = 'bytes'
        else:
            array_a = np.zeros(n_pcs, dtype=np.uint8) if filename_a else None
            array_b = np.zeros(n_pcs, dtype=np.uint8) if filename_b else None
            output_type = 'array'

        # Loop
        while self.acquiring.is_set():

            # Define n, the ammount of bytes to be generated
            if n_remaining // n_chunk:
                n = n_chunk
            else:
                n = n_remaining

            # Generate n pcs
            pcs_a, pcs_b = rng.get_rng_pulse_counts(n, output_type=output_type, timeout=None)

            # Test for RNG disconnection
            if (pcs_a is None) or (pcs_b is None):
                self.acquiring.clear()
                self.cbkreg_finished(None)
                continue

            # Write pcs
            if out_file:
                if rng_out == 'A':
                    f_a.write(pcs_a)

                elif rng_out == 'B':
                    f_b.write(pcs_b)

                elif rng_out == 'AB':
                    f_a.write(pcs_a)
                    f_b.write(pcs_b)

            else:
                if rng_out == 'A':
                    array_a[n_acquired:n_acquired+n] = pcs_a

                elif rng_out == 'B':
                    array_b[n_acquired:n_acquired+n] = pcs_b

                elif rng_out == 'AB':
                    array_a[n_acquired:n_acquired+n] = pcs_a
                    array_b[n_acquired:n_acquired+n] = pcs_b

            # Update n_acquired
            n_acquired += n

            # Progress callback
            progress = n_acquired / n_pcs * 100
            try:
                self.cbkfcn_progress(progress)
            except Exception as e:
                self.lg.error('{} Pulse Counts: Error calling cbkfcn_progress -- {}'.format(self.name, e))

            # Finished? Quit loop
            n_remaining = n_pcs - n_acquired

            if n_remaining == 0:
                self.acquiring.clear()

        # Output
        if out_file:
            f_a.close() if f_a else None
            f_b.close() if f_b else None
            outputs = [f for f in [filename_a, filename_b] if f is not None]
            outputs_str = ', '.join(outputs)
        else:
            outputs = [a for a in [array_a, array_b] if a is not None]
            outputs_str = ['array' for a in outputs]

        # Delete incomplete?
        if self.delete_incomplete and (progress < 100):
            if out_file:
                [os.remove(f) for f in outputs]
            else:
                for a in outputs:
                    del a
            outputs = []
            outputs_str = ['None']

        # Elapsed time
        time_s = time.perf_counter() - time_start
        time_td = datetime.timedelta(seconds=time_s)
        time_str = '{}'.format(str(time_td).split('.')[0])

        # Print info
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Pulse Counts: Finished {} acquisition with device {}\n{}Progress = {:.0f}%, Time={}\n{}Generated: {}'
            .format(self.name, out_type, rng.dev_serial_number, LOG_FILL, progress, time_str, LOG_FILL, outputs_str))

        # Return data
        return (outputs, progress, time_str)


    ## BYTES

    def get_filenames_bytes(self, rng, n_bytes, postproc, rng_out, out_path):
        # RNG serial number
        sn = rng.dev_serial_number

        # Postproc ID
        postproc_id = D_RNG_POSTPROC[postproc]

        # Filenames base
        ammount_str = get_ammount_prefix_str(n_bytes)
        filename_a = '{}_BYTES_{}_PP{}_A__'.format(sn, ammount_str, postproc_id)
        filename_b = '{}_BYTES_{}_PP{}_B__'.format(sn, ammount_str, postproc_id)
        filename_abxor = '{}_BYTES_{}_PP{}_AB_XOR__'.format(sn, ammount_str, postproc_id)
        filename_abalt = '{}_BYTES_{}_PP{}_AB_ALT__'.format(sn, ammount_str, postproc_id)

        # Filenames count
        n_filename_a = len(glob(os.path.join(out_path, filename_a + '*.bin')))
        n_filename_b = len(glob(os.path.join(out_path, filename_b + '*.bin')))
        n_filename_ab = max(n_filename_a, n_filename_b)
        n_filename_abxor = len(glob(os.path.join(out_path, filename_abxor + '*.bin')))
        n_filename_abalt = len(glob(os.path.join(out_path, filename_abalt + '*.bin')))

        # Result Filename: count corrected + path
        if rng_out == 'A':
            filename_a = os.path.join(out_path, '{}{}.bin'.format(filename_a, n_filename_a + 1))
            return filename_a, None

        elif rng_out == 'B':
            filename_b = os.path.join(out_path, '{}{}.bin'.format(filename_b, n_filename_b + 1))
            return None, filename_b

        elif rng_out == 'AB':
            filename_a = os.path.join(out_path, '{}{}.bin'.format(filename_a, n_filename_ab + 1))
            filename_b = os.path.join(out_path, '{}{}.bin'.format(filename_b, n_filename_ab + 1))
            return filename_a, filename_b

        elif rng_out == 'AB_XOR':
            filename_ab = os.path.join(out_path, '{}{}.bin'.format(filename_abxor, n_filename_abxor + 1))
            return filename_ab, None

        elif rng_out == 'AB_ALT':
            filename_ab = os.path.join(out_path, '{}{}.bin'.format(filename_abalt, n_filename_abalt + 1))
            return filename_ab, None


    def get_bytes(self, rng, n_bytes, n_chunk=10000, postproc='NONE', rng_out='AB', out_file=False, out_path='',
                  threaded=False):
        # Test function inputs
        if not rng.connected():
            self.lg.error('{} Bytes: Trying to acquire data from a disconnected RNG'.format(self.name))
            return [], 0, ''

        if self.acquiring.is_set():
            self.lg.error('{} Bytes: RNG is already acquiring'.format(self.name))
            return [], 0, ''

        if out_file and not os.path.isdir(out_path):
            self.lg.error('{} Bytes: The provided out_path doesn\'t exists'.format(self.name))
            return [], 0, ''

        if n_bytes <= 0:
            self.lg.error('{} Bytes: Provide n_bytes > 0'.format(self.name))
            return [], 0, ''

        if n_chunk <= 0:
            self.lg.error('{} Bytes: Provide n_chunk > 0'.format(self.name))
            return [], 0, ''

        if postproc not in D_RNG_POSTPROC.keys():
            self.lg.error('{} Bytes: Unknown postproc. Please choose from {}'.format(self.name, D_RNG_POSTPROC.keys()))
            return [], 0, ''

        if rng_out not in ACQ_BYTES_RNG_OUT:
            self.lg.error('{} Bytes: Unknown rng_out. Please choose from {}'.format(self.name, ACQ_BYTES_RNG_OUT))
            return [], 0, ''

        # Start generation
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Bytes: Starting {} acquisition with device {}'.format(self.name, out_type, rng.dev_serial_number))

        args = (rng, n_bytes, n_chunk, postproc, rng_out, out_file, out_path)

        if threaded:
            result_future = self.thread_executor.submit(self.loop_bytes, *args)
            return result_future
        else:
            result = self.loop_bytes(*args)
            return result


    def loop_bytes(self, rng, n_bytes, n_chunk, postproc, rng_out, out_file, out_path):
        # Set variables
        time_start = time.perf_counter()
        postproc_id = D_RNG_POSTPROC[postproc]
        n_acquired = 0
        n_remaining = n_bytes
        progress = 0.
        self.acquiring.set()

        # Output
        out_path = '' if not out_file else out_path
        filename_a, filename_b = self.get_filenames_bytes(rng, n_bytes, postproc, rng_out, out_path)

        if out_file:
            f_a = open(filename_a, mode='bw') if filename_a else None
            f_b = open(filename_b, mode='bw') if filename_b else None
            output_type = 'bytes'
        else:
            array_a = np.zeros(n_bytes, dtype=np.uint8) if filename_a else None
            array_b = np.zeros(n_bytes, dtype=np.uint8) if filename_b else None
            output_type = 'array'

        # Loop
        while self.acquiring.is_set():

            # Define n, the ammount of bytes to be generated
            if n_remaining // n_chunk:
                n = n_chunk
            else:
                n = n_remaining

            if rng_out == 'AB_ALT':
                n = n // 2

            # Generate n bytes
            bytes_a, bytes_b = rng.get_rng_bytes(n, postproc_id, output_type=output_type, timeout=None)

            # Test for RNG disconnection
            if (bytes_a is None) or (bytes_b is None):
                self.acquiring.clear()
                self.cbkreg_finished(None)
                continue

            # Write bytes
            if out_file:
                if rng_out == 'A':
                    f_a.write(bytes_a)

                elif rng_out == 'B':
                    f_b.write(bytes_b)

                elif rng_out == 'AB':
                    f_a.write(bytes_a)
                    f_b.write(bytes_b)

                elif rng_out == 'AB_XOR':
                    int_a = int.from_bytes(bytes_a, 'little')
                    int_b = int.from_bytes(bytes_b, 'little')
                    int_abxor = int_a ^ int_b
                    bytes_ab = int_abxor.to_bytes(len(bytes_a), 'little')
                    f_a.write(bytes_ab)

                elif rng_out == 'AB_ALT': # Alternate rng A and B bytes
                    bytes_ab = bytes(itertools.chain.from_iterable(zip(bytes_a, bytes_b)))
                    f_a.write(bytes_ab)

            else:
                if rng_out == 'A':
                    array_a[n_acquired:n_acquired+n] = bytes_a

                elif rng_out == 'B':
                    array_b[n_acquired:n_acquired+n] = bytes_b

                elif rng_out == 'AB':
                    array_a[n_acquired:n_acquired+n] = bytes_a
                    array_b[n_acquired:n_acquired+n] = bytes_b

                elif rng_out == 'AB_XOR':
                    bytes_ab = np.bitwise_xor(bytes_a, bytes_b)
                    array_a[n_acquired:n_acquired+n] = bytes_ab

                elif rng_out == 'AB_ALT': # Alternate rng A and B bytes
                    bytes_ab = list(itertools.chain.from_iterable(zip(bytes_a, bytes_b)))
                    array_a[n_acquired:n_acquired+2*n] = bytes_ab

            # Update n_acquired
            if rng_out == 'AB_ALT':
                n_acquired += 2*n
            else:
                n_acquired += n

            # Progress callback
            progress = n_acquired / n_bytes * 100
            try:
                self.cbkfcn_progress(progress)
            except Exception as e:
                self.lg.error('{} Bytes: Error calling cbkfcn_progress -- {}'.format(self.name, e))

            # Finished? Quit loop
            n_remaining = n_bytes - n_acquired

            if n_remaining == 0:
                self.acquiring.clear()

        # Output
        if out_file:
            f_a.close() if f_a else None
            f_b.close() if f_b else None
            outputs = [f for f in [filename_a, filename_b] if f is not None]
            outputs_str = ', '.join(outputs)
        else:
            outputs = [a for a in [array_a, array_b] if a is not None]
            outputs_str = ['array' for a in outputs]

        # Delete incomplete?
        if self.delete_incomplete and (progress < 100):
            if out_file:
                [os.remove(f) for f in outputs]
            else:
                for a in outputs:
                    del a
            outputs = []
            outputs_str = ['None']

        # Elapsed time
        time_s = time.perf_counter() - time_start
        time_td = datetime.timedelta(seconds=time_s)
        time_str = '{}'.format(str(time_td).split('.')[0])

        # Print info
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Bytes: Finished {} acquisition with device {}\n{}Progress = {:.0f}%, Time={}\n{}Generated: {}'
                .format(self.name, out_type, rng.dev_serial_number, LOG_FILL, progress, time_str, LOG_FILL, outputs_str))

        # Return data
        return outputs, progress, time_str


    ## NUMBERS

    def get_filenames_numbers(self, rng, out_file_binary, out_path, n_nums, num_type):
        # RNG serial number
        sn = rng.dev_serial_number

        # Extension
        ext = 'npy' if out_file_binary else 'dat'

        # Filenames base
        num_type_str = 'INT' if num_type is int else 'FLOAT'
        ammount_str = get_ammount_prefix_str(n_nums)
        filename = '{}_{}_{}__'.format(sn, num_type_str, ammount_str)

        # Filenames count
        n_filename = len(glob(os.path.join(out_path, filename + '*.{}'.format(ext))))

        # Result Filename: count corrected + path
        return os.path.join(out_path, '{}{}.{}'.format(filename, n_filename + 1, ext))


    def get_numbers(self, rng, n_nums, n_chunk=5000, num_type=float, num_min=0, num_max=1,
                    out_file=False, out_path='', out_file_binary=True, out_file_separator='\n', threaded=False):
        # Test function inputs
        if not rng.connected():
            self.lg.error('{} Numbers: Trying to acquire data from a disconnected RNG'.format(self.name))
            return [], 0, ''

        if self.acquiring.is_set():
            self.lg.error('{} Numbers: RNG is already acquiring'.format(self.name))
            return [], 0, ''

        if out_file and not os.path.isdir(out_path):
            self.lg.error('{} Numbers: The provided out_path doesn\'t exists'.format(self.name))
            return [], 0, ''

        if n_nums <= 0:
            self.lg.error('{} Numbers: Provide n_nums > 0'.format(self.name))
            return [], 0, ''

        if n_chunk <= 0:
            self.lg.error('{} Numbers: Provide n_chunk > 0'.format(self.name))
            return [], 0, ''

        if num_type not in [int, float]:
            self.lg.error('{} Numbers: Unknown num_type. Please choose from [int, float]'.format(self.name))
            return [], 0, ''

        if num_type is int:
            num_delta = num_max - num_min
            if num_delta >= 65536:
                self.lg.error('{} Numbers: Provide num_max - num_min < 65536'.format(self.name))
                return [], 0, ''

        if out_file_separator not in [',', ';', '.', '\n']:
            self.lg.error('{} Numbers: Unknown out_file_separator. Please choose from {}'
                     .format(self.name, [',', ';', '.', '\n']))
            return [], 0, ''

        # Start generation
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Numbers: Starting {} acquisition with device {}'.format(self.name, out_type, rng.dev_serial_number))

        args = (rng, n_nums, n_chunk, num_type, num_min, num_max, out_file, out_path, out_file_binary, out_file_separator)

        if threaded:
            result_future = self.thread_executor.submit(self.loop_numbers, *args)
            return result_future
        else:
            result = self.loop_numbers(*args)
            return result


    def loop_numbers(self, rng, n_nums, n_chunk, num_type, num_min, num_max, out_file, out_path, out_file_binary, out_file_separator):
        # Set variables
        time_start = time.perf_counter()
        n_acquired = 0
        n_remaining = n_nums
        progress = 0.
        self.acquiring.set()
        num_delta = num_max - num_min

        # Output dtype
        if num_type is int:
            if num_min + num_delta < 2**8:
                dtype_out = np.uint8
            elif num_min + num_delta < 2**16:
                dtype_out = np.uint16
            elif num_min + num_delta < 2**32:
                dtype_out = np.uint32
        else:
            dtype_out = np.float32

        # Output
        if out_file and out_file_binary == False:
            filename = self.get_filenames_numbers(rng, out_file_binary, out_path, n_nums, num_type)
            f = open(filename, mode='w')
        elif out_file and out_file_binary == True:
            filename = self.get_filenames_numbers(rng, out_file_binary, out_path, n_nums, num_type)
            array = np.zeros(n_nums, dtype=dtype_out)
        else:
            array = np.zeros(n_nums, dtype=dtype_out)

        # Loop
        while self.acquiring.is_set():

            # Define n, the ammount of bytes to be generated
            if n_remaining // n_chunk:
                n = n_chunk
            else:
                n = n_remaining

            # Generate n numbers
            if num_type is int and num_delta < 256:
                nums = rng.get_rng_int8s(n, int(num_delta), output_type='array', timeout=None)
            elif num_type is int and num_delta < 65536:
                nums = rng.get_rng_int16s(n, int(num_delta), output_type='array', timeout=None)
            else:
                nums = rng.get_rng_floats(n, output_type='array', timeout=None)

            # Test for RNG disconnection
            if nums is None:
                self.acquiring.clear()
                self.cbkreg_finished(None)
                continue

            # Correct range
            if num_type is int:
                nums = int(num_min) + nums.astype(dtype_out)
            else:
                nums = num_min + nums * num_delta

            # Write numbers
            if out_file and out_file_binary == False:
                nums.tofile(f, sep=out_file_separator)
                if n == n_chunk:
                    f.write(out_file_separator)
            else:
                array[n_acquired:n_acquired+n] = nums

            # Update n_acquired
            n_acquired += n

            # Progress callback
            progress = n_acquired / n_nums * 100
            try:
                self.cbkfcn_progress(progress)
            except Exception as e:
                self.lg.warning('{} Numbers: Error calling cbkfcn_progress -- {}'.format(self.name, e))

            # Finished? Quit loop
            n_remaining = n_nums - n_acquired

            if n_remaining == 0:
                self.acquiring.clear()

        # Output
        if out_file and out_file_binary == False:
            f.close()

            outputs = [filename]
            outputs_str = ', '.join(outputs)
        elif out_file and out_file_binary == True:
            np.save(filename, array)

            outputs = [filename]
            outputs_str = ', '.join(outputs)
        else:
            outputs = [array]
            outputs_str = ['array' for a in outputs]

        # Delete incomplete?
        if self.delete_incomplete and (progress < 100):
            if out_file:
                [os.remove(filename) for f in outputs]
            else:
                for a in outputs:
                    del a
            outputs = []
            outputs_str = ['None']

        # Elapsed time
        time_s = time.perf_counter() - time_start
        time_td = datetime.timedelta(seconds=time_s)
        time_str = '{}'.format(str(time_td).split('.')[0])

        # Print info
        out_type = 'file' if out_file else 'array'
        self.lg.info('{} Numbers: Finished {} acquisition with device {}\n{}Progress = {:.0f}%, Time={}\n{}Generated: {}'
                .format(self.name, out_type, rng.dev_serial_number, LOG_FILL, progress, time_str, LOG_FILL, outputs_str))

        # Return data
        return outputs, progress, time_str


"""
Copyright (c) 2024 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

"""
Progress window used by the Acquisition sub-app.
"""

import tkinter as tk
from tkinter import ttk
import tkinter.messagebox as tkm


### VARS

PAD = 10


### WIN_PROGRESS

class WIN_PROGRESS(tk.Toplevel):

    def __init__(self, parent):
        # Initiate WIN_PROGRESS
        super().__init__(parent)
        self.name = 'WIN_PROGRESS'
        self.title(' Progress ')
        self.geometry('350x100+{}+{}'.format(self.master.winfo_x() + self.master.winfo_width() + 10, self.master.winfo_y()))
        self.rowconfigure([0], weight=1)
        self.columnconfigure([0], weight=1)
        self.resizable(False, False) # No resize
        self.protocol('WM_DELETE_WINDOW', lambda: None)

        # Widgets
        self.frame = ttk.Frame(self, padding=(PAD, PAD))
        self.frame.grid(row=0, column=0, sticky='nsew')
        self.frame.columnconfigure([0], weight=1)
        self.frame.rowconfigure([0,1], weight=1)

        self.var_progress = tk.DoubleVar(value=0)
        self.pb_progress = ttk.Progressbar(self.frame, orient=tk.HORIZONTAL, length=200, mode='determinate', variable=self.var_progress, maximum=100.0)
        self.pb_progress.grid(row=0, column=0, sticky='ew', padx=2*PAD)

        self.bt_cancel = ttk.Button(self.frame, text='Cancel', command=self.prog_cancel)
        self.bt_cancel.grid(row=1, column=0)

        # Start
        self.acquiring = False

    
    def set_extra_title(self, extra_str=''):
        self.title(' Progress ' + extra_str)


    def show(self):
        self.acquiring = True
        self.prog_update(0)
        self.geometry('350x100+{}+{}'.format(self.master.winfo_x() + self.master.winfo_width() + 10, self.master.winfo_y()))
        self.wm_deiconify()
        self.master.update()    # Update widgets on non-threaded mode
        self.grab_set()         # Can't perform other actions while progress window is shown


    def hide(self):
        self.acquiring = False
        self.grab_release()
        self.withdraw()
        self.set_extra_title('')
    

    def __enter__(self):
        self.show()


    def __exit__(self, exc_type, exc_val, exc_tb):
        self.hide()


    def prog_update(self, progress):
        self.var_progress.set(progress)
        self.master.update()    # Update widgets on non-threaded mode


    def prog_cancel(self):        
        if tkm.askyesno(parent=self, title=('Cancel'), message='Cancel?'):
            self.acquiring = False
            try:
                self.master.rng_acq.stop()
            except:
                pass


"""
Copyright (c) 2023 Gabriel Guerrer

Distributed under the MIT license - See LICENSE for details
"""

from .win_progress import WIN_PROGRESS
from .rava_subapp_acquisition import RAVA_SUBAPP_ACQUISITION, rava_subapp_acquisition